import re
import math
import logging
import pybedtools

logger = logging.getLogger(__name__)

def write_general_assembly_stats(refobj, queryobj, contigregions, gapregions, outputfiles, benchparams, args)->dict:

    bmstats = {}

    generalstatspath = outputfiles["generalstatsfile"]
    scafflengthfile = outputfiles["scaffoldlengths"]
    contiglengthfile = outputfiles["contiglengths"]

    # total bases on benchmark haplotypes (for NG/LG calculation):
    hap1totalbases = 0
    hap2totalbases = 0
    matpattern = benchparams['matpattern']
    patpattern = benchparams['patpattern']
    phap1 = re.compile(r".*" + re.escape(matpattern) + ".*", re.IGNORECASE)
    phap2 = re.compile(r".*" + re.escape(patpattern) + ".*", re.IGNORECASE)
    for refcontig in refobj.references:
        hap1match = phap1.match(refcontig)
        hap2match = phap2.match(refcontig)
        logger.debug("Looking for mat/pat class of " + refcontig + " with strings " + matpattern + ", " + patpattern)
        if hap1match:
            hap1totalbases = hap1totalbases + refobj.get_reference_length(refcontig)
            logger.debug(refcontig + " matches mat pattern")
        if hap2match:
            hap2totalbases = hap2totalbases + refobj.get_reference_length(refcontig)
            logger.debug(refcontig + " matches pat pattern")

    bmstats['hap1totalbases'] = hap1totalbases # this is the total number of bases in the MATERNAL benchmark chroms
    bmstats['hap2totalbases'] = hap2totalbases # this is the total number of bases in the PATERNAL benchmark chroms
    diploidtotalbases = hap1totalbases + hap2totalbases # this is the total number of bases in the entire diploid benchmark
    bmstats['diploidtotalbases'] = diploidtotalbases
    avghaptotalbases = (hap1totalbases + hap2totalbases)/2.0
    bmstats['totalbases'] = hap1totalbases + hap2totalbases

    mincontiglength = args.mincontiglength
    with open(generalstatspath, "w") as gsfh:
        numscaffolds = queryobj.nreferences
        totalscaffoldbases = sum(queryobj.lengths)

        cumscaffbases = 0
        numscaffs = 0
        scaffold_n50 = 0
        scaffold_ng50 = 0
        scaffold_l50 = 0
        scaffold_lg50 = 0
        scaffold_aung = 0
        scaffold_lengths = queryobj.lengths

        scaffold_lengths.sort(reverse=True)
        with open(scafflengthfile, "w") as sfh:
            for scafflength in scaffold_lengths:
                cumscaffbases = cumscaffbases + scafflength
                numscaffs = numscaffs + 1
                if cumscaffbases > 0.5*totalscaffoldbases and scaffold_n50 == 0:
                    scaffold_n50 = scafflength
                    scaffold_l50 = numscaffs
                if cumscaffbases > 0.5*diploidtotalbases and scaffold_ng50 == 0:
                    scaffold_ng50 = scafflength
                    scaffold_lg50 = numscaffs
                if diploidtotalbases > 0:
                    scaffold_aung = scaffold_aung + scafflength*scafflength/diploidtotalbases
                perctotallength = int(1000.0*cumscaffbases/diploidtotalbases + 0.5)/10.0
                # At some point, may want to pull scaffold name into last field for use in annotating plots
                sfh.write(str(perctotallength) + "\t" + str(scafflength) + "\t" + str(cumscaffbases) + "\tNA\n")

        totalns = 0
        for gap in gapregions:
            gaplength = len(gap)
            totalns = totalns + gaplength

        numcontigs = len(contigregions)
        contig_n50 = 0
        contig_ng50 = 0
        contig_l50 = 0
        contig_lg50 = 0
        contig_aung = 0
        numlargecontigs = 0
        sizelist = []
        totalsize = 0
        for contig in contigregions:
            contiglength = len(contig)
            if contiglength >= mincontiglength:
                numlargecontigs = numlargecontigs + 1
                sizelist.append(contiglength)
                totalsize = totalsize + contiglength

        sizelist.sort(reverse=True)
        cumcontigbases = 0
        numcontigs = 0
        with open(contiglengthfile, "w") as cfh:
            for contigsize in sizelist:
                cumcontigbases = cumcontigbases + contigsize
                numcontigs = numcontigs + 1
                if cumcontigbases > 0.5*totalsize and contig_n50 == 0:
                    contig_n50 = contigsize
                    contig_l50 = numcontigs
                if cumcontigbases > 0.5*diploidtotalbases and contig_ng50 == 0:
                    contig_ng50 = contigsize
                    contig_lg50 = numcontigs
                if diploidtotalbases > 0:
                    contig_aung = contig_aung + contigsize*contigsize/diploidtotalbases
                perctotallength = int(1000.0*cumcontigbases/diploidtotalbases + 0.5)/10.0
                cfh.write(str(perctotallength) + "\t" + str(contigsize) + "\t" + str(cumcontigbases) + "\tNA\n")
        
        bmstats['totalns'] = totalns

        bmstats['numscaffolds'] = numscaffolds
        bmstats['totalscaffoldbases'] = totalscaffoldbases
        bmstats['scaffoldn50'] = scaffold_n50
        bmstats['scaffoldng50'] = scaffold_ng50
        bmstats['scaffoldl50'] = scaffold_l50
        bmstats['scaffoldlg50'] = scaffold_lg50
        bmstats['scaffoldaung'] = scaffold_aung

        bmstats['numcontigs'] = numcontigs
        bmstats['numlargecontigs'] = numcontigs
        bmstats['totallargecontigbases'] = totalsize
        bmstats['contign50'] = contig_n50
        bmstats['contigng50'] = contig_ng50
        bmstats['contigl50'] = contig_l50
        bmstats['contiglg50'] = contig_lg50
        bmstats['contigaung'] = contig_aung

        gsfh.write("Scaffolds:\n\n")
        gsfh.write("Number of scaffolds: " + str(numscaffolds) + "\n")
        gsfh.write("Total bases in scaffolds: " + str(totalscaffoldbases) + "\n")
        gsfh.write("Ns per kb in scaffolds: " + str(int(totalns*1000/totalscaffoldbases)) + "\n")
        gsfh.write("Scaffold N50: " + str(round(scaffold_n50/1000000, 3)) + "Mb\n")
        gsfh.write("Scaffold NG50: ")
        gsfh.write(str(round(scaffold_ng50/1000000, 3)) + "Mb\n")
        gsfh.write("Scaffold L50: " + str(scaffold_l50) + "\n")
        gsfh.write("Scaffold LG50: ")
        gsfh.write(str(scaffold_lg50) + "\n")
        gsfh.write("Scaffold auNG: ")
        gsfh.write(str(round(scaffold_aung/1000000, 3)) + "Mb\n")
        gsfh.write("\nContigs:\n\n")
        gsfh.write("Number of contigs: " + str(numcontigs) + "\n")
        gsfh.write("Number of contigs >= " + str(mincontiglength) + ": " + str(numlargecontigs) + "\n")
        gsfh.write("Total bases in contigs >= " + str(mincontiglength) + ": " + str(totalsize) + "\n")
        gsfh.write("Contig N50: " + str(round(contig_n50/1000000, 3)) + "Mb\n")
        gsfh.write("Contig NG50: ")
        gsfh.write(str(round(contig_ng50/1000000, 3)) + "Mb\n")
        gsfh.write("Contig L50: " + str(contig_l50) + "\n")
        gsfh.write("Contig LG50: ")
        gsfh.write(str(contig_lg50) + "\n")
        gsfh.write("Contig auNG: ")
        gsfh.write(str(round(contig_aung/1000000, 3)) + "Mb\n")
        gsfh.write("\n")

    return bmstats

def write_merged_aligned_stats(refobj, queryobj, mergedtruthcoveredbed, includedtruthcoveredbed, mergedtestmatcoveredbed, mergedtestpatcoveredbed, bedfiles:dict, bmstats:dict, benchparams, args)->dict:

    generalstatsfile = bedfiles["generalstatsfile"]

    matpattern = benchparams['matpattern']
    patpattern = benchparams['patpattern']
    phap1 = re.compile(r".*" + re.escape(matpattern) + ".*", re.IGNORECASE)
    phap2 = re.compile(r".*" + re.escape(patpattern) + ".*", re.IGNORECASE)
    totalbenchcovered = 0
    totalincludedbenchcovered = 0

    totalrefaligned = 0
    numberrefaligns = 0

    totalbases = bmstats['diploidtotalbases'] # total bases in benchmark

    if totalbases > 0:
        nga50 = 0
        lga50 = 0
        nga90 = 0
        lga90 = 0
        aunga = 0

    for truthint in sorted(mergedtruthcoveredbed, key=lambda h: len(h), reverse=True):
        alignlength = len(truthint)
        totalrefaligned = totalrefaligned + alignlength
        numberrefaligns = numberrefaligns + 1

        if totalbases > 0:
            if totalrefaligned >= 0.5*totalbases and nga50 == 0:
                nga50 = alignlength
                lga50 = numberrefaligns
            if totalrefaligned >= 0.9*totalbases and nga90 == 0:
                nga90 = alignlength
                lga90 = numberrefaligns
            aunga = aunga + alignlength*alignlength/totalbases

    matbenchcovered = 0
    patbenchcovered = 0
    for truthint in mergedtruthcoveredbed:
        [chrom, start, end, name] = truthint
        chrom = truthint.chrom
        start = int(truthint.start)
        end = int(truthint.stop)
        totalbenchcovered = totalbenchcovered + end - start
        if phap1.match(chrom):
            matbenchcovered = matbenchcovered + end - start
        if phap2.match(chrom):
            patbenchcovered = patbenchcovered + end - start

    for truthint in includedtruthcoveredbed:
        [chrom, start, end, name] = truthint
        chrom = truthint.chrom
        start = int(truthint.start)
        end = int(truthint.stop)
        totalincludedbenchcovered = totalincludedbenchcovered + end - start

    longesttestalignment = 0
    totaltestmatcovered = 0
    totaltestpatcovered = 0
    for testint in mergedtestmatcoveredbed:
        alignlength = len(testint)
        totaltestmatcovered = totaltestmatcovered + alignlength
        if alignlength > longesttestalignment:
            longesttestalignment = alignlength

    for testint in mergedtestpatcoveredbed:
        alignlength = len(testint)
        totaltestpatcovered = totaltestpatcovered + alignlength
        if alignlength > longesttestalignment:
            longesttestalignment = alignlength

    bmstats["testmattotalcovered"] = totaltestmatcovered
    bmstats["testpattotalcovered"] = totaltestpatcovered
    bmstats["benchtotalcovered"] = totalbenchcovered
    bmstats["includedbenchtotalcovered"] = totalincludedbenchcovered

    with open(generalstatsfile, "a") as gsfh:
        gsfh.write("Aligned contig bases:\n\n")
        if totalbases > 0:
            gsfh.write("NGA50: " + str(round(nga50/1000000, 3)) + "Mb\n")
            gsfh.write("LGA50: " + str(lga50) + "\n")
            gsfh.write("NGA90: " + str(round(nga90/1000000, 3)) + "Mb\n")
            gsfh.write("LGA90: " + str(lga90) + "\n")
            gsfh.write("auNGA: " + str(round(aunga/1000000, 3)) + "Mb\n")
        perctestmatcovered = int(totaltestmatcovered * 1000 / bmstats['totallargecontigbases'] + 0.5) / 10
        perctestpatcovered = int(totaltestpatcovered * 1000 / bmstats['totallargecontigbases'] + 0.5) / 10
        if bmstats['hap1totalbases'] > 0 or bmstats['hap2totalbases'] > 0:
            percbenchcovered = int(1000 * totalbenchcovered / (bmstats['hap1totalbases'] + bmstats['hap2totalbases']) + 0.5) / 10
        else:
            percbenchcovered = 'NA'
        if bmstats['hap1totalbases'] > 0:
            percmatbenchcovered = int(1000 * matbenchcovered / bmstats['hap1totalbases'] + 0.5) / 10
        else:
            percmatbenchcovered = 'NA'
        if bmstats['hap2totalbases'] > 0:
            percpatbenchcovered = int(1000 * patbenchcovered / bmstats['hap2totalbases'] + 0.5) / 10
        else:
            percpatbenchcovered = 'NA'
        gsfh.write("Total " + args.assembly + " bases in aligns to MAT chromosomes: " + str(totaltestmatcovered) + "/" + str(bmstats['totallargecontigbases']) + " (" + str(perctestmatcovered) + "% of total bases in contigs >= " + str(args.mincontiglength) + " bases)" + "\n")
        gsfh.write("Total " + args.assembly + " bases in aligns to PAT chromosomes: " + str(totaltestpatcovered) + "/" + str(bmstats['totallargecontigbases']) + " (" + str(perctestpatcovered) + "% of total bases in contigs >= " + str(args.mincontiglength) + " bases)" + "\n")
        gsfh.write("Longest " + args.assembly + " alignment length (in test assembly bases) to the benchmark: " + str(round(longesttestalignment/1000000, 3)) + "Mb\n")
        gsfh.write("Total " + args.benchmark + " covered: " + str(totalbenchcovered) + "/" + str(bmstats['totalbases']) + " (" + str(percbenchcovered) + "% of all MAT and PAT bases in benchmark)" + "\n" )
        gsfh.write("MAT " + args.benchmark + " covered: " + str(matbenchcovered) + "/" + str(bmstats['hap1totalbases']) + " (" + str(percmatbenchcovered) + "% of MAT bases in benchmark)" + "\n")
        gsfh.write("PAT " + args.benchmark + " covered: " + str(patbenchcovered) + "/" + str(bmstats['hap2totalbases']) + " (" + str(percpatbenchcovered) + "% of PAT bases in benchmark)" + "\n")

        if "ngc50" in bmstats.keys():
            maxclusterdistance = args.maxclusterdistance
            gsfh.write("\nAlignment cluster length stats (alignments are clustered to combine collinear alignments with ref and query shifts less than " + str(maxclusterdistance) + " bases:\n\n")
            gsfh.write("NGC50: " + str(round(bmstats['ngc50']/1000000, 3)) + "Mb\n")
            gsfh.write("LGC50: " + str(bmstats['lgc50']) + "\n")
            gsfh.write("NGC90: " + str(round(bmstats['ngc90']/1000000, 3)) + "Mb\n")
            gsfh.write("LGC90: " + str(bmstats['lgc90']) + "\n")
            gsfh.write("auNGC: " + str(round(bmstats['aungc']/1000000, 3)) + "Mb\n")

    return bmstats

def write_aligned_cluster_stats(outputfiles:dict, bmstats:dict, args)->int:
    structstatsfile = outputfiles['clusterlengths']

    alignclusters = bmstats['alignclusters']
    hap1totalbases = bmstats['hap1totalbases'] # total MATERNAL bases in benchmark
    hap2totalbases = bmstats['hap2totalbases'] # total PATERNAL bases in benchmark
    avghaptotalbases = (hap1totalbases + hap2totalbases)/2.0
    diploidtotalbases = hap1totalbases + hap2totalbases

    allclusters = []

    for refentry in alignclusters.keys():
        refclusters = alignclusters[refentry]
        for cluster in refclusters:
            clusteraligns = cluster["aligns"]
            clusterlow = None
            clusterhigh = None
            for align in clusteraligns:
                if clusterlow is None or align["targetstart"] < clusterlow:
                    clusterlow = align["targetstart"]
                if clusterhigh is None or align["targetend"] > clusterhigh:
                    clusterhigh = align["targetend"]

            logger.debug(refentry + " cluster " + str(clusterlow) + "-" + str(clusterhigh))
            cluster["spanlength"] = clusterhigh - clusterlow + 1
            cluster["refentry"] = refentry
            allclusters.append(cluster)

    allclusters.sort(key=lambda x:x["spanlength"], reverse=True)
    totallength = 0
    with open(structstatsfile, "w") as cfh:
        for cluster in allclusters:
            spanlength = cluster["spanlength"]
            totallength = totallength + spanlength
            perctotallength = int(1000.0*totallength/diploidtotalbases + 0.5)/10.0
            refentry = cluster["refentry"]
            logger.debug(str(spanlength) + "\t" + str(totallength) + "\t" + refentry)
            cfh.write(str(perctotallength) + "\t" + str(spanlength) + "\t" + str(totallength) + "\t" + refentry + "\n")

    if totallength > 0:
        ngc50 = 0
        lgc50 = 0
        ngc90 = 0
        lgc90 = 0
        aungc = 0
        numberclusters = 0
        totalclusterbases = 0

        for cluster in allclusters:
            spanlength = cluster["spanlength"]
            numberclusters = numberclusters + 1
            totalclusterbases = totalclusterbases + spanlength

            if totalclusterbases >= 0.5*totallength and ngc50 == 0:
                ngc50 = spanlength
                lgc50 = numberclusters
            if totalclusterbases >= 0.9*totallength and ngc90 == 0:
                ngc90 = spanlength
                lgc90 = numberclusters
            aungc = aungc + spanlength*spanlength/totallength
        bmstats['ngc50'] = ngc50
        bmstats['lgc50'] = lgc50
        bmstats['ngc90'] = ngc90
        bmstats['lgc90'] = lgc90
        bmstats['aungc'] = aungc

    return 0

def write_aligned_stats(refobj, queryobj, truthcoveredbed, bedfiles:dict, bmstats:dict, benchparams, args)->dict:

    generalstatsfile = bedfiles["generalstatsfile"]

    matpattern = benchparams['matpattern']
    patpattern = benchparams['patpattern']
    phap1 = re.compile(r".*" + re.escape(matpattern) + ".*", re.IGNORECASE)
    phap2 = re.compile(r".*" + re.escape(patpattern) + ".*", re.IGNORECASE)
    totalbenchcovered = 0

    totalrefaligned = 0
    numberrefaligns = 0

    totalbases = bmstats['diploidtotalbases'] # total bases in benchmark

    nga50 = 0
    lga50 = 0
    nga90 = 0
    lga90 = 0
    aunga = 0
    for truthint in sorted(truthcoveredbed, key=lambda h: len(h), reverse=True):
        alignlength = len(truthint)
        totalrefaligned = totalrefaligned + alignlength
        numberrefaligns = numberrefaligns + 1

        if totalrefaligned >= 0.5*totalbases and nga50 == 0:
            nga50 = alignlength
            lga50 = numberrefaligns
        aunga = aunga + alignlength*alignlength/totalbases

    matbenchcovered = 0
    patbenchcovered = 0
    for truthint in truthcoveredbed:
        [chrom, start, end, name] = truthint
        chrom = truthint.chrom
        start = int(truthint.start)
        end = int(truthint.stop)
        totalbenchcovered = totalbenchcovered + end - start
        if phap1.match(chrom):
            matbenchcovered = matbenchcovered + end - start
        if phap2.match(chrom):
            patbenchcovered = patbenchcovered + end - start

    bmstats["benchtotalcovered"] = totalbenchcovered

    with open(generalstatsfile, "a") as gsfh:
        gsfh.write("Aligned contig bases:\n\n")
        gsfh.write("NGA50: " + str(round(nga50/1000000, 3)) + "Mb\n")
        gsfh.write("LGA50: " + str(lga50) + "\n")
        gsfh.write("NGA90: " + str(round(nga90/1000000, 3)) + "Mb\n")
        gsfh.write("LGA90: " + str(lga90) + "\n")
        gsfh.write("auNGA: " + str(round(aunga/1000000, 3)) + "Mb\n")
        if bmstats['diploidtotalbases'] > 0 or bmstats['diploidtotalbases'] > 0:
            percbenchcovered = int(1000 * totalbenchcovered / bmstats['diploidtotalbases'] + 0.5) / 10
        else:
            percbenchcovered = 'NA'
        if bmstats['hap1totalbases'] > 0:
            percmatbenchcovered = int(1000 * matbenchcovered / bmstats['hap1totalbases'] + 0.5) / 10
        else:
            percmatbenchcovered = 'NA'
        if bmstats['hap2totalbases'] > 0:
            percpatbenchcovered = int(1000 * patbenchcovered / bmstats['hap2totalbases'] + 0.5) / 10
        else:
            percpatbenchcovered = 'NA'
        gsfh.write("Total " + args.benchmark + " covered: " + str(totalbenchcovered) + "/" + str(bmstats['totalbases']) + " (" + str(percbenchcovered) + "% of all MAT and PAT bases in benchmark)" + "\n" )
        gsfh.write("MAT " + args.benchmark + " covered: " + str(matbenchcovered) + "/" + str(bmstats['hap1totalbases']) + " (" + str(percmatbenchcovered) + "% of MAT bases in benchmark)" + "\n")
        gsfh.write("PAT " + args.benchmark + " covered: " + str(patbenchcovered) + "/" + str(bmstats['hap2totalbases']) + " (" + str(percpatbenchcovered) + "% of PAT bases in benchmark)" + "\n")

    return bmstats

def write_qv_stats(benchmark_stats:dict, alignedscorecounts:list, snverrorscorecounts:list, indelerrorscorecounts:list, bedfiles:dict, args):

    variantfile = bedfiles["bencherrortypebed"]
    totalerrors = 0
    totalindelerrors = 0
    totalsnverrors = 0
    totalphasingerrors = 0
    totalphasingindelerrors = 0
    totalphasingsnverrors = 0
    totalconsensuserrors = 0
    totalconsensusindelerrors = 0
    totalconsensussnverrors = 0

    with open(variantfile, "r") as vfh:
        errorline = vfh.readline()
        while errorline:
            errorline = errorline.rstrip()
            [chrom, start, end, name, score, strand, widestart, wideend, color, errortype, vartype, varname] = errorline.split("\t")
            namefields = name.split("_")
            refallele = namefields[-2]
            altallele = namefields[-1]
            refallelelength = len(refallele) # these lengths are wrong when allele is "*"--should replace
            altallelelength = len(altallele) # these lengths are wrong when allele is "*"--should replace

            totalerrors = totalerrors + 1
            if vartype == 'INDEL':
                totalindelerrors = totalindelerrors + 1
            else:
                totalsnverrors = totalsnverrors + 1
            if errortype == "PHASING":
                totalphasingerrors = totalphasingerrors + 1
                if refallele == "*" or altallele == "*" or refallelelength != altallelelength:
                    totalphasingindelerrors = totalphasingindelerrors + 1
                else:
                    totalphasingsnverrors = totalphasingsnverrors + 1
            elif errortype == "CONSENSUS":
                totalconsensuserrors = totalconsensuserrors + 1
                if vartype == 'INDEL':
                    totalconsensusindelerrors = totalconsensusindelerrors + 1
                else:
                    totalconsensussnverrors = totalconsensussnverrors + 1

            errorline = vfh.readline()

    totalassemblybasesinaligns = benchmark_stats["includedbenchtotalcovered"]
    if totalassemblybasesinaligns > 0:
        phaseerrorrate = totalphasingerrors / totalassemblybasesinaligns
        phasesnverrorrate = totalphasingsnverrors / totalassemblybasesinaligns
        phaseindelerrorrate = totalphasingindelerrors / totalassemblybasesinaligns
        consensuserrorrate = totalconsensuserrors / totalassemblybasesinaligns
        consensussnverrorrate = totalconsensussnverrors / totalassemblybasesinaligns
        consensusindelerrorrate = totalconsensusindelerrors / totalassemblybasesinaligns
    else:
        phaseerrorrate = 'NA'
        phasesnverrorrate = 'NA'
        phaseindelerrorrate = 'NA'
        consensuserrorrate = 'NA'
        consensussnverrorrate = 'NA'
        consensusindelerrorrate = 'NA'

    if consensuserrorrate != "NA":
        if consensuserrorrate > 0:
            consensusqv = int(-10 * math.log(consensuserrorrate, 10) + 0.5)
        else:
            consensusqv = 'Inf'

        if consensuserrorrate + phaseerrorrate > 0:
            qvwithphaseerrors = int(-10 * math.log(consensuserrorrate + phaseerrorrate, 10) + 0.5)
        else:
            qvwithphaseerrors = 'Inf'
    else:
        consensusqv = 'NA'
        qvwithphaseerrors = 'NA'

    benchmark_stats["assemblyqv"] = consensusqv

    if consensussnverrorrate != "NA":
        if consensussnverrorrate > 0:
            consensussnvqv = int(-10 * math.log(consensussnverrorrate, 10) + 0.5)
        else:
            consensussnvqv = 'Inf'

        if consensussnverrorrate + phasesnverrorrate > 0:
            qvwithsnvphaseerrors = int(-10 * math.log(consensussnverrorrate + phasesnverrorrate, 10) + 0.5)
        else:
            qvwithsnvphaseerrors = 'Inf'
    else:
        consensussnvqv = 'NA'
        qvwithsnvphaseerrors = 'NA'

    benchmark_stats["assemblysnvqv"] = consensussnvqv

    if consensusindelerrorrate != "NA":
        if consensusindelerrorrate > 0:
            consensusindelqv = int(-10 * math.log(consensusindelerrorrate, 10) + 0.5)
        else:
            consensusindelqv = 'Inf'

        if consensusindelerrorrate + phaseindelerrorrate > 0:
            qvwithindelphaseerrors = int(-10 * math.log(consensusindelerrorrate + phaseindelerrorrate, 10) + 0.5)
        else:
            qvwithindelphaseerrors = 'Inf'
    else:
        consensusindelqv = 'NA'
        qvwithindelphaseerrors = 'NA'

    benchmark_stats["assemblyindelqv"] = consensusindelqv

    snvtypestring = ""
    totaltypesnverrors = 0
    if totalassemblybasesinaligns > 0:
        for errortype in sorted(benchmark_stats["singlebasecounts"].keys()):
            typeerrors = benchmark_stats["singlebasecounts"][errortype]
            snvtypeerrorspermb = typeerrors/totalassemblybasesinaligns*1000000
            snvtypestring = snvtypestring + errortype + "\t" + str(typeerrors) + "\t" + str(snvtypeerrorspermb) + "\n"
            totaltypesnverrors = totaltypesnverrors + typeerrors
        snverrorspermb = int(totaltypesnverrors/totalassemblybasesinaligns*1000000)
    
        with open(bedfiles["snvstatsfile"], "w") as sfh:
            sfh.write(snvtypestring)
    else:
        logger.info("Not writing statistics regarding substitutions because no bases aligned!")

    if totalassemblybasesinaligns > 0:
        indellengthstring = ""
        totalhistindelerrors = 0
        for indellength in sorted(benchmark_stats["indellengthcounts"].keys()):
            indelcount = benchmark_stats["indellengthcounts"][indellength]
            indelerrorspermb = indelcount/totalassemblybasesinaligns*1000000
            indellengthstring = indellengthstring + str(indellength) + "\t" + str(indelcount) + "\t" + str(indelerrorspermb) + "\n"
            totalhistindelerrors = totalhistindelerrors + indelcount
        indelerrorspermb = int(totalhistindelerrors/totalassemblybasesinaligns*1000000)
    
        with open(bedfiles["indelstatsfile"], "w") as ifh:
            ifh.write(indellengthstring)
    else:
        logger.info("Not writing statistics regarding indels because no bases aligned!")

    generalstatsfile = bedfiles["generalstatsfile"]
    with open(generalstatsfile, "a") as gsfh:
        gsfh.write("\nAssembly accuracy:\n\n")
        gsfh.write("Total errors in alignments of " + args.assembly + " to " + args.benchmark + ": " + str(totalerrors) + "\n")
        gsfh.write("Total substitution errors in alignments of " + args.assembly + " to " + args.benchmark + ": " + str(totalsnverrors) + "\n")
        gsfh.write("Total indel errors in alignments of " + args.assembly + " to " + args.benchmark + ": " + str(totalindelerrors) + "\n")
        gsfh.write("Errors where sequence matches the opposite haplotype: " + str(totalphasingerrors) + " (" + str(totalphasingsnverrors) + " subs. and " + str(totalphasingindelerrors) + " indels)" + "\n")
        gsfh.write("Errors where sequence doesn\'t match the opposite haplotype: " + str(totalconsensuserrors) + " (" + str(totalconsensussnverrors) + " subs. and " + str(totalconsensusindelerrors) + " indels)" + "\n")
        gsfh.write("Number of included, non-excluded benchmark bases with primary-aligned assembly scaffold sequence: " + str(totalassemblybasesinaligns) + "\n")
        gsfh.write("Overall QV without wrong-haplotype errors: " + str(consensusqv) + "\n")
        gsfh.write("Overall QV including wrong-haplotype errors: " + str(qvwithphaseerrors) + "\n")
        gsfh.write("Overall SNV QV without wrong-haplotype errors: " + str(consensussnvqv) + "\n")
        gsfh.write("Overall SNV QV including wrong-haplotype errors: " + str(qvwithsnvphaseerrors) + "\n")
        gsfh.write("Overall Indel QV without wrong-haplotype errors: " + str(consensusindelqv) + "\n")
        gsfh.write("Overall Indel QV including wrong-haplotype errors: " + str(qvwithindelphaseerrors) + "\n")

    if len(alignedscorecounts) > 0:
        with open(bedfiles["qvstatsfile"], "w") as qfh:
            for i in range(len(alignedscorecounts)):
                if i > 0 or alignedscorecounts[i] > 0:
                    alignedbases = alignedscorecounts[i]
                    snverrorbases = snverrorscorecounts[i]
                    indelerrorbases = indelerrorscorecounts[i]
                    if alignedbases > 0:
                        consensuserrorrate = (snverrorbases + indelerrorbases)/alignedbases
                        if consensuserrorrate > 0:
                            binqvscore = str(int(-10 * math.log(consensuserrorrate, 10) + 0.5))
                        else:
                            binqvscore = "Inf"
                    else:
                        binqvscore = "NA"
                    qfh.write(str(i) + "\t" + str(snverrorbases) + "\t" + str(indelerrorbases) + "\t" + str(alignedbases) + "\t" + binqvscore + "\n")


    generalstatsfile = bedfiles["generalstatsfile"]
    
    return 0

def write_mononuc_stats(mononucstats:dict, bedfiles:dict, benchmark_stats:dict, args):

    totalcoveredmononucs = len(mononucstats.keys())
    totalcorrect = 0
    totalwronglength = 0
    totalphaseswitches = 0
    totalcomplex = 0

    for variantname in mononucstats.keys():
        mononucdict = mononucstats[variantname]
        repeatedbase = mononucdict["base"]
        varianttype = mononucdict["type"]
        reflength = mononucdict["length"]
        newlength = mononucdict["assemblylength"]
        if newlength == -1:
            totalcomplex = totalcomplex + 1
        elif varianttype == "CORRECT":
            totalcorrect = totalcorrect + 1
        elif varianttype == "PHASING":
            totalphaseswitches = totalphaseswitches + 1
        elif varianttype == "CONSENSUS":
            totalwronglength = totalwronglength + 1

    generalstatsfile = bedfiles["generalstatsfile"]
    with open(generalstatsfile, "a") as gsfh:
        gsfh.write("\nMononucleotide run accuracy:\n\n")
        gsfh.write("Total number of true assembly mononucleotide runs of ten or more bases covered by " + args.assembly + " alignments: " + str(totalcoveredmononucs) + "\n")
        if totalcoveredmononucs > 0:
            perccorrect = round(100*totalcorrect/totalcoveredmononucs, 3)
        else:
            perccorrect = 'NA'
        gsfh.write("Number of mononucleotide runs correct in assembly: " + str(totalcorrect) + " (" + str(perccorrect) + "%)" + "\n")
        if totalcoveredmononucs > 0:
            percphaseswitch = round(100*totalphaseswitches/totalcoveredmononucs, 3)
        else:
            percphaseswitch = 'NA'
        gsfh.write("Number of mononucleotide runs with assembly alleles matching the alternate haplotype: " + str(totalphaseswitches) + " (" + str(percphaseswitch) + "%)" + "\n")
        if totalcoveredmononucs > 0:
            percwronglength = round(100*totalwronglength/totalcoveredmononucs, 3)
        else:
            percwronglength = 'NA'
        gsfh.write("Number of mononucleotide runs with non-alternate alleles of fewer or more of the same base in the assembly: " + str(totalwronglength) + " (" + str(percwronglength) + "%)" + "\n")
        if totalcoveredmononucs > 0:
            perccomplex = round(100*totalcomplex/totalcoveredmononucs, 3)
        else:
            perccomplex = 'NA'
        gsfh.write("Number of mononucleotide runs with erroneous alleles other than extensions or contractions: " + str(totalcomplex) + " (" + str(perccomplex) + "%)" + "\n")

    return 0

def write_het_stats(bedfiles:dict, bmstats:dict, benchparams, args):

    hetbedfile = bedfiles["coveredhetsitealleles"]
    num_switches = 0
    num_maternal = 0
    num_paternal = 0
    total_matching_hets = 0

    matpattern = benchparams['matpattern']
    patpattern = benchparams['patpattern']
    pmat = re.compile(r".*" + re.escape(matpattern) + ".*", re.IGNORECASE)
    ppat = re.compile(r".*" + re.escape(patpattern) + ".*", re.IGNORECASE)

    last_contig = ""
    last_hap = ""
    with open(hetbedfile, "r") as hfh:
        hetline = hfh.readline()
        while hetline:
            hetline = hetline.rstrip()
            [contig, start, end, name, allele, chrom, chromstart, chromend, widecontig,  widestart, wideend, phasetype] = hetline.split("\t")
            alignedhap = "NA"
            if pmat.match(chrom):
                alignedhap = "MAT"
            elif ppat.match(chrom):
                alignedhap = "PAT"
            else:
                logger.warning("Uncertain haplotype for benchmark chromosome " + chrom)
            curallele_hap = "NA"
            if (phasetype == "SAMEHAP" and alignedhap == "MAT") or (phasetype == "ALTHAP" and alignedhap == "PAT"):
                num_maternal = num_maternal + 1
                total_matching_hets = total_matching_hets + 1
                curallele_hap = "MAT"
            if (phasetype == "SAMEHAP" and alignedhap == "PAT") or (phasetype == "ALTHAP" and alignedhap == "MAT"):
                num_paternal = num_paternal + 1
                curallele_hap = "PAT"
                total_matching_hets = total_matching_hets + 1

            # record switch if we're on the same contig and the haplotype changed:
            if contig == last_contig and curallele_hap != last_hap:
                num_switches = num_switches + 1
            last_contig = contig
            last_hap = curallele_hap

            hetline = hfh.readline()


    bmstats["numhetswitches"] = num_switches
    bmstats["nummaternalhetalleles"] = num_maternal
    bmstats["numpaternalhetalleles"] = num_paternal
    bmstats["phaseswitchespermb"] = num_switches/bmstats["totallargecontigbases"]*1000000

    generalstatsfile = bedfiles["generalstatsfile"]
    with open(generalstatsfile, "a") as gsfh:
        gsfh.write("\nPhasing statistics (" + str(total_matching_hets) + " benchmark heterozygous sites with an aligned maternal or paternal assembly allele):\n\n")
        gsfh.write("Number of maternal alleles within alignments: " + str(num_maternal) + "\n")
        gsfh.write("Number of paternal alleles within alignments: " + str(num_paternal) + "\n")
        gsfh.write("Number of hap switches at heterozygous site alleles within alignments: " + str(num_switches) + "\n")
        gsfh.write("Het site switch rate per megabase within aligned sequence: " + str(bmstats["phaseswitchespermb"]) + " per Mb\n")

    return 0

def write_hmm_phaseswitch_stats(outputfiles:dict, bmstats:dict, args):

    phasebedfile = outputfiles["phaseblockbed"]
    num_switches = 0
    num_scaffbases = 0

    last_scaff = ""
    last_hap = ""
    with open(phasebedfile, "r") as pfh:
        phaseblockline = pfh.readline()
        while phaseblockline:
            phaseblockline = phaseblockline.rstrip()
            fields = phaseblockline.split("\t")
            scaff = fields[0]
            start = int(fields[1])
            end = int(fields[2])
            hap = fields[3]

            if scaff == last_scaff and hap != last_hap:
                num_switches = num_switches + 1
            else:
                last_scaff = scaff
                last_hap = hap

            num_scaffbases = num_scaffbases + end - start
            phaseblockline = pfh.readline()

    bmstats["numhmmphaseswitches"] = num_switches
    bmstats["numscaffoldbases"] = num_scaffbases
    bmstats["hmmphaseswitchespermb"] = num_switches/num_scaffbases*1000000

    generalstatsfile = outputfiles["generalstatsfile"]
    with open(generalstatsfile, "a") as gsfh:
        gsfh.write("Number of within-scaffold switches in HMM-predicted haplotype: " + str(num_switches) + "\n")
        gsfh.write("HMM phase switch rate per megabase within scaffold sequences: " + str(bmstats["hmmphaseswitchespermb"]) + " per Mb\n")

    return 0

def write_read_mononuc_stats(stats:dict, outputfiles:dict, args):

    histogramfile = outputfiles["mononuchistfile"]
    lengthcomposition = {}
    with open(histogramfile, "w") as msfh:
        msfh.write("BenchmarkLength\tReadLength\tNumCorrect\tNumHetAllele\tNumError\tNumComplex\n")
        for mononuclength in sorted(stats.keys()):
            lengthcomposition[mononuclength] = {'CORRECT':0, 'HET':0, 'INSERROR':0, 'DELERROR':0, 'COMPLEX':0, 'FLANKERROR':0}
            for readmononuclength in sorted(stats[mononuclength].keys()):
                numcorrect = 0
                if "CORRECT" in stats[mononuclength][readmononuclength].keys():
                    numcorrect = stats[mononuclength][readmononuclength]["CORRECT"]
                    lengthcomposition[mononuclength]['CORRECT'] = lengthcomposition[mononuclength]['CORRECT'] + numcorrect
                numhetallele = 0
                if "HET" in stats[mononuclength][readmononuclength].keys():
                    numhetallele = stats[mononuclength][readmononuclength]["HET"]
                    lengthcomposition[mononuclength]['HET'] = lengthcomposition[mononuclength]['HET'] + numhetallele
                numerror = 0
                if "LENGTHERROR" in stats[mononuclength][readmononuclength].keys():
                    numerror = stats[mononuclength][readmononuclength]["LENGTHERROR"]
                    if readmononuclength > mononuclength:
                        lengthcomposition[mononuclength]['INSERROR'] = lengthcomposition[mononuclength]['INSERROR'] + numerror
                    elif readmononuclength < mononuclength:
                        lengthcomposition[mononuclength]['DELERROR'] = lengthcomposition[mononuclength]['DELERROR'] + numerror
                numcomplexallele = 0
                if "COMPLEX" in stats[mononuclength][readmononuclength].keys():
                    numcomplexallele = stats[mononuclength][readmononuclength]["COMPLEX"]
                    lengthcomposition[mononuclength]['COMPLEX'] = lengthcomposition[mononuclength]['COMPLEX'] + numcomplexallele
                if "FLANKERROR" in stats[mononuclength][readmononuclength].keys():
                    numflankallele = stats[mononuclength][readmononuclength]["FLANKERROR"]
                    lengthcomposition[mononuclength]['FLANKERROR'] = lengthcomposition[mononuclength]['FLANKERROR'] + numflankallele
                msfh.write(str(mononuclength) + "\t" + str(readmononuclength) + "\t" + str(numcorrect) + "\t" + str(numhetallele) + "\t" + str(numerror) + "\t" + str(numcomplexallele) + "\n")

    mononuccompfile = outputfiles["mononuccompositionfile"]
    mononucoverviewfile = outputfiles["mononucoverviewfile"]
    totalcorrect = 0
    totalerror = 0
    totalcomplex = 0
    totalinsertions = 0
    totaldeletions = 0
    totalflankerror = 0
    with open(mononuccompfile, "w") as mcfh:
        mcfh.write("BenchmarkLength\tNumCorrect\tNumHetAllele\tNumInsError\tNumDelError\tNumComplex\n")
        for mononuclength in sorted(stats.keys()):
            mcfh.write(str(mononuclength) + "\t" + str(lengthcomposition[mononuclength]['CORRECT']) + "\t" + str(lengthcomposition[mononuclength]['HET']) + "\t" + str(lengthcomposition[mononuclength]['INSERROR']) + "\t" + str(lengthcomposition[mononuclength]['DELERROR']) + "\t" + str(lengthcomposition[mononuclength]['COMPLEX']) + "\n")
            totalcorrect = totalcorrect + lengthcomposition[mononuclength]['CORRECT']
            totalerror = totalerror + lengthcomposition[mononuclength]['INSERROR'] + lengthcomposition[mononuclength]['DELERROR']
            totalcomplex = totalcomplex + lengthcomposition[mononuclength]['COMPLEX']
            totalflankerror = totalflankerror + lengthcomposition[mononuclength]['FLANKERROR']
            totalinsertions = totalinsertions + lengthcomposition[mononuclength]['INSERROR']
            totaldeletions = totaldeletions + lengthcomposition[mononuclength]['DELERROR']

    with open(mononucoverviewfile, "w") as mofh:
        if totalcorrect + totalerror > 0:
            noncomplexaccuracy = str(int(totalcorrect/(totalcorrect + totalerror)*1000)/10) + "%"
        else:
            noncomplexaccuracy = "NA"
        if totalcorrect + totalerror + totalcomplex > 0:
            accuracywithcomplex = str(int(totalcorrect/(totalcorrect + totalerror + totalcomplex)*1000)/10) + "%"
        else:
            accuracywithcomplex = "NA"
        if totalcorrect + totalerror + totalflankerror > 0:
            accuracywithflankerror = str(int(totalcorrect/(totalcorrect + totalerror + totalflankerror)*1000)/10) + "%"
        else:
            accuracywithflankerror = "NA"
        if totalcorrect + totalerror + totalcomplex + totalflankerror > 0:
            accuracywithcomplexandflankerror = str(int(totalcorrect/(totalcorrect + totalerror + totalcomplex + totalflankerror)*1000)/10) + "%"
        else:
            accuracywithcomplexandflankerror = "NA"

        if totaldeletions > 0:
            overcallundercallratio = str(int(100*totalinsertions/totaldeletions)/100)
        else:
            overcallundercallratio = "NA"
        mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (not counting read alleles that match the alternate haplotype or are complex): " + noncomplexaccuracy + "\n")
        mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (including complex): " + accuracywithcomplex + "\n")
        mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (including flankerror): " + accuracywithflankerror + "\n")
        mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (including complex and flankerror): " + accuracywithcomplexandflankerror + "\n")
        mofh.write("Ratio of overcalled to undercalled homopolymer runs: " + overcallundercallratio + "\n")
            
def write_read_str_stats(strlabel:str, stats:dict, outputfiles:dict, args):

    lengthaccuracyfile = outputfiles["strlengthaccuracyprefix"] + "." + strlabel + ".txt"
    lengthdiffaccfile = outputfiles["strlengthdiffaccuracyprefix"] + "." + strlabel + ".txt"

    correctbylength = {}
    complexbylength = {}
    flankbylength = {}
    lengtherrorbylength = {}
    hetbylength = {}
   
    # tally length diff reads
    lengthdiffreads = {}
    complexreads = 0
    flankerrorreads = 0
    hetreads = 0
    
    for refstrlength in sorted(stats.keys()):
        correctbylength[refstrlength] = 0
        lengtherrorbylength[refstrlength] = 0
        complexbylength[refstrlength] = 0
        flankbylength[refstrlength] = 0
        hetbylength[refstrlength] = 0
        for readstrlength in sorted(stats[refstrlength].keys()):
            if readstrlength != -1:
                lengthdiff = readstrlength - refstrlength
                if lengthdiff not in lengthdiffreads.keys():
                    lengthdiffreads[lengthdiff] = 0
            else:
                lengthdiff = -1
            if "CORRECT" in stats[refstrlength][readstrlength].keys():
                correctbylength[refstrlength] = correctbylength[refstrlength] + stats[refstrlength][readstrlength]['CORRECT']
                lengthdiffreads[lengthdiff] = lengthdiffreads[lengthdiff] + stats[refstrlength][readstrlength]['CORRECT']
            if "LENGTHERROR" in stats[refstrlength][readstrlength].keys():
                lengtherrorbylength[refstrlength] = lengtherrorbylength[refstrlength] + stats[refstrlength][readstrlength]['LENGTHERROR']
                lengthdiffreads[lengthdiff] = lengthdiffreads[lengthdiff] + stats[refstrlength][readstrlength]['LENGTHERROR']
            if "COMPLEX" in stats[refstrlength][readstrlength].keys():
                complexbylength[refstrlength] = complexbylength[refstrlength] + stats[refstrlength][readstrlength]['COMPLEX']
                complexreads = complexreads + stats[refstrlength][readstrlength]['COMPLEX']
            if "FLANKERROR" in stats[refstrlength][readstrlength].keys():
                flankbylength[refstrlength] = flankbylength[refstrlength] + stats[refstrlength][readstrlength]['FLANKERROR']
                flankerrorreads = flankerrorreads + stats[refstrlength][readstrlength]['FLANKERROR']
            if "HET" in stats[refstrlength][readstrlength].keys():
                hetbylength[refstrlength] = hetbylength[refstrlength] + stats[refstrlength][readstrlength]['HET']
                hetreads = hetreads + stats[refstrlength][readstrlength]['HET']

    with open(lengthaccuracyfile, "w") as lfh:
        lfh.write("RefLength\tCorrect\tLengthError\tComplexError\tFlankError\tHet\n")
        for refstrlength in sorted(correctbylength.keys()):
            accrate = correctbylength[refstrlength]/(correctbylength[refstrlength] + lengtherrorbylength[refstrlength] + complexbylength[refstrlength] + flankbylength[refstrlength])
            lfh.write(str(refstrlength) + "\t" + str(correctbylength[refstrlength]) + "\t" + str(lengtherrorbylength[refstrlength]) + "\t" + str(complexbylength[refstrlength]) + "\t" + str(flankbylength[refstrlength]) + "\t" + str(hetbylength[refstrlength]) + "\t" + str(accrate) + "\n")

    with open(lengthdiffaccfile, "w") as dfh:
        dfh.write("LengthDiff\tType\tNumReads\n")
        for lengthdiff in sorted(lengthdiffreads.keys()):
            if lengthdiff == 0:
                dfh.write(str(lengthdiff) + "\tCORRECT\t" + str(lengthdiffreads[lengthdiff]) + "\n")
            else:
                dfh.write(str(lengthdiff) + "\tLENGTHERROR\t" + str(lengthdiffreads[lengthdiff]) + "\n")
        dfh.write("0\tCOMPLEX\t" + str(complexreads) + "\n")
        dfh.write("0\tFLANKERROR\t" + str(flankerrorreads) + "\n")
        dfh.write("0\tHET\t" + str(hetreads) + "\n")

    #lengthcomposition = {}
    #with open(histogramfile, "w") as sfh:
        #sfh.write("BenchmarkLength\tReadLength\tNumCorrect\tNumHetAllele\tNumError\tNumComplex\n")
        #for mononuclength in sorted(stats.keys()):
            #lengthcomposition[mononuclength] = {'CORRECT':0, 'HET':0, 'INSERROR':0, 'DELERROR':0, 'COMPLEX':0, 'FLANKERROR':0}
            #for readmononuclength in sorted(stats[mononuclength].keys()):
                #numcorrect = 0
                #if "CORRECT" in stats[mononuclength][readmononuclength].keys():
                    #numcorrect = stats[mononuclength][readmononuclength]["CORRECT"]
                    #lengthcomposition[mononuclength]['CORRECT'] = lengthcomposition[mononuclength]['CORRECT'] + numcorrect
                #numhetallele = 0
                #if "HET" in stats[mononuclength][readmononuclength].keys():
                    #numhetallele = stats[mononuclength][readmononuclength]["HET"]
                    #lengthcomposition[mononuclength]['HET'] = lengthcomposition[mononuclength]['HET'] + numhetallele
                #numerror = 0
                #if "LENGTHERROR" in stats[mononuclength][readmononuclength].keys():
                    #numerror = stats[mononuclength][readmononuclength]["LENGTHERROR"]
                    #if readmononuclength > mononuclength:
                        #lengthcomposition[mononuclength]['INSERROR'] = lengthcomposition[mononuclength]['INSERROR'] + numerror
                    #elif readmononuclength < mononuclength:
                        #lengthcomposition[mononuclength]['DELERROR'] = lengthcomposition[mononuclength]['DELERROR'] + numerror
                #numcomplexallele = 0
                #if "COMPLEX" in stats[mononuclength][readmononuclength].keys():
                    #numcomplexallele = stats[mononuclength][readmononuclength]["COMPLEX"]
                    #lengthcomposition[mononuclength]['COMPLEX'] = lengthcomposition[mononuclength]['COMPLEX'] + numcomplexallele
                #if "FLANKERROR" in stats[mononuclength][readmononuclength].keys():
                    #numflankallele = stats[mononuclength][readmononuclength]["FLANKERROR"]
                    #lengthcomposition[mononuclength]['FLANKERROR'] = lengthcomposition[mononuclength]['FLANKERROR'] + numflankallele
                #sfh.write(str(mononuclength) + "\t" + str(readmononuclength) + "\t" + str(numcorrect) + "\t" + str(numhetallele) + "\t" + str(numerror) + "\t" + str(numcomplexallele) + "\n")
#
    #mononuccompfile = outputfiles["mononuccompositionfile"]
    #mononucoverviewfile = outputfiles["mononucoverviewfile"]
    #totalcorrect = 0
    #totalerror = 0
    #totalcomplex = 0
    #totalinsertions = 0
    #totaldeletions = 0
    #totalflankerror = 0
    #with open(mononuccompfile, "w") as mcfh:
        #mcfh.write("BenchmarkLength\tNumCorrect\tNumHetAllele\tNumInsError\tNumDelError\tNumComplex\n")
        #for mononuclength in sorted(stats.keys()):
            #mcfh.write(str(mononuclength) + "\t" + str(lengthcomposition[mononuclength]['CORRECT']) + "\t" + str(lengthcomposition[mononuclength]['HET']) + "\t" + str(lengthcomposition[mononuclength]['INSERROR']) + "\t" + str(lengthcomposition[mononuclength]['DELERROR']) + "\t" + str(lengthcomposition[mononuclength]['COMPLEX']) + "\n")
            #totalcorrect = totalcorrect + lengthcomposition[mononuclength]['CORRECT']
            #totalerror = totalerror + lengthcomposition[mononuclength]['INSERROR'] + lengthcomposition[mononuclength]['DELERROR']
            #totalcomplex = totalcomplex + lengthcomposition[mononuclength]['COMPLEX']
            #totalflankerror = totalflankerror + lengthcomposition[mononuclength]['FLANKERROR']
            #totalinsertions = totalinsertions + lengthcomposition[mononuclength]['INSERROR']
            #totaldeletions = totaldeletions + lengthcomposition[mononuclength]['DELERROR']
#
    #with open(mononucoverviewfile, "w") as mofh:
        #if totalcorrect + totalerror > 0:
            #noncomplexaccuracy = str(int(totalcorrect/(totalcorrect + totalerror)*1000)/10) + "%"
        #else:
            #noncomplexaccuracy = "NA"
        #if totalcorrect + totalerror + totalcomplex > 0:
            #accuracywithcomplex = str(int(totalcorrect/(totalcorrect + totalerror + totalcomplex)*1000)/10) + "%"
        #else:
            #accuracywithcomplex = "NA"
        #if totalcorrect + totalerror + totalflankerror > 0:
            #accuracywithflankerror = str(int(totalcorrect/(totalcorrect + totalerror + totalflankerror)*1000)/10) + "%"
        #else:
            #accuracywithflankerror = "NA"
        #if totalcorrect + totalerror + totalcomplex + totalflankerror > 0:
            #accuracywithcomplexandflankerror = str(int(totalcorrect/(totalcorrect + totalerror + totalcomplex + totalflankerror)*1000)/10) + "%"
        #else:
            #accuracywithcomplexandflankerror = "NA"
#
        #if totaldeletions > 0:
            #overcallundercallratio = str(int(100*totalinsertions/totaldeletions)/100)
        #else:
            #overcallundercallratio = "NA"
        #mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (not counting read alleles that match the alternate haplotype or are complex): " + noncomplexaccuracy + "\n")
        #mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (including complex): " + accuracywithcomplex + "\n")
        #mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (including flankerror): " + accuracywithflankerror + "\n")
        #mofh.write("Percent accuracy of homopolymer runs of 10 or more bases (including complex and flankerror): " + accuracywithcomplexandflankerror + "\n")
        #mofh.write("Ratio of overcalled to undercalled homopolymer runs: " + overcallundercallratio + "\n")

def write_read_error_summary(stats:dict, outputfiles:dict):
    totalsnverrors = 0
    totalindelerrors = 0
    snvtypestring = ""
    indellengthstring = ""
    with open(outputfiles["errorstatsfile"], "w") as efh:
        efh.write("Total aligned bases: " + str(stats["totalalignedbases"]) + "\n")
        efh.write("Total clipped bases: " + str(stats["totalclippedbases"]) + "\n")
        errorspermb = int(stats["totalerrorsinaligns"]/stats["totalalignedbases"]*1000000)
        efh.write("Total errors in alignments: " + str(stats["totalerrorsinaligns"]) + " (" + str(errorspermb) + " error per megabase)" + "\n")
        for errortype in sorted(stats["singlebasecounts"].keys()):
            typeerrors = stats["singlebasecounts"][errortype]
            snvtypeerrorspermb = typeerrors/stats["totalalignedbases"]*1000000
            snvtypestring = snvtypestring + errortype + "\t" + str(typeerrors) + "\t" + str(snvtypeerrorspermb) + "\n"
            totalsnverrors = totalsnverrors + typeerrors
        snverrorspermb = int(totalsnverrors/stats["totalalignedbases"]*1000000)
        efh.write("Total substitution errors in alignments: " + str(totalsnverrors) + " (" + str(snverrorspermb) + " error per megabase)" + "\n")
        for indellength in sorted(stats["indellengthcounts"].keys()):
            indelcount = stats["indellengthcounts"][indellength]
            indelerrorspermb = indelcount/stats["totalalignedbases"]*1000000
            indellengthstring = indellengthstring + str(indellength) + "\t" + str(indelcount) + "\t" + str(indelerrorspermb) + "\n"
            totalindelerrors = totalindelerrors + indelcount
        indelerrorspermb = int(totalindelerrors/stats["totalalignedbases"]*1000000)
        efh.write("Total indel errors in alignments: " + str(totalindelerrors) + " (" + str(indelerrorspermb) + " error per megabase)" + "\n")

    with open(outputfiles["snvstatsfile"], "w") as sfh:
        sfh.write(snvtypestring)

    with open(outputfiles["indelstatsfile"], "w") as ifh:
        ifh.write(indellengthstring)

    if len(stats["alignedqualscorecounts"]) > 0:
        with open(outputfiles["qvstatsfile"], "w") as qfh:
            for i in range(len(stats["alignedqualscorecounts"])):
                alignedbases = stats["alignedqualscorecounts"][i]
                snverrorbases = stats["snverrorqualscorecounts"][i]
                indelerrorbases = stats["indelerrorqualscorecounts"][i]
                qfh.write(str(i) + "\t" + str(snverrorbases) + "\t" + str(indelerrorbases) + "\t" + str(alignedbases) + "\n")
    #if len(stats["positiontotalcounts"]) > 0:
        #with open(outputfiles["positionstatsfile"], "w") as pfh:
            ## i is the zero-based position in the read
            #for i in range(len(stats["positionsnvcounts"])):
                #alignedbases = stats["positiontotalcounts"][i]
                #indelerrorbases = stats["positionindelcounts"][i]
                #pfh.write(str(i+1) + "\t" + str(indelerrorbases) + "\tIndel\t" + str(alignedbases) + "\n")
                #for snvkey in stats["positionsnvcounts"][i].keys():
                    #snverrorbases = stats["positionsnvcounts"][i]
                    #pfh.write(str(i+1) + "\t" + str(snverrorbases) + "\t" + snvkey + "\t" + str(alignedbases) + "\n")

    return 0

def write_phase_block_stats(phaseblockints:pybedtools.bedtool.BedTool, outputfiles:dict, benchmark_stats:dict, args):

    return 0

def write_assembly_haplotype_stats(fastaobj, contigregions, gapregions, args)->dict:

    hapstats = {}

    return hapstats

def write_assembly_comparison_stats(args):

    return 0




import re
import pysam
import pybedtools
import logging
from collections import namedtuple
from GQC import seqparse
from GQC import phasing

logger = logging.getLogger(__name__)

# This routine, ported from mummer's delta-filter C++ routines "flagQLIS" and "flagRLIS", finds the longest
# increasing sequence of alignments along a target- or query-sorted list of alignments, using the
# method's "sorttype" argument to determine whether to do QLIS (sorttype="query") or RLIS (sorttype=
# "target").
#
# Inputs: The list of alignments passed to the routine should contain a dictionary for each align
# with fields: target, targetstart, targetend, targetlength, query, querystart, queryend, querylength
# For reverse strand alignments, querystart can be larger than queryend, but targetstart should not 
# be larger than targetend. All coordinates are assumed to be 1-based.
#
# Return value: The function returns a list with the same alignments passed as input, except the list 
# doesn't contain any of the aligns that are part of the QLIS or RLIS, and it is sorted in order of
# increasing query or reference lower-coordinate.
#

def filter_aligns(alignlist:list, sorttype="target", maxoverlap=0.95, mummeralgorithm=True):

    # define the fields that will be used in sorting and determination of length, etc.
    if sorttype == "target":
        sortfield = "targetstart"
        lowfield = "targetstart"
        highfield = "targetend"
    else:
        sortfield = "querylow"
        lowfield = "querylow"
        highfield = "queryhigh"
   
    for align in alignlist:
        # set query high/low
        if align["querystart"] < align["queryend"]:
            align["querylow"] = align["querystart"]
            align["queryhigh"] = align["queryend"]
        else:
            align["querylow"] = align["queryend"]
            align["queryhigh"] = align["querystart"]
    
    # make a hash grouping the alignments by target or query entry
    alignbyentrydict = create_align_by_entry_dict(alignlist, entrytype=sorttype)

    # create a hash with keys for each alignment that passes the LIS filter
    filteredaligns = []
    for entry in alignbyentrydict:
        alignlist = sorted(alignbyentrydict[entry], key=lambda align: align[sortfield])
        numaligns = len(alignlist)
        lis = []

        if mummeralgorithm:
            for i in range(numaligns):
                lis.append({"used":False})
            allbest = []
    
            while True:
                for i in range(numaligns):
                    if lis[i]["used"]:
                        continue
                    if "identity" not in alignlist[i]:
                        alignlist[i]["identity"] = 1.0
                    lis[i]["align"] = alignlist[i]
                    lis[i]["score"] = (alignlist[i][highfield] - alignlist[i][lowfield] + 1)*alignlist[i]["identity"]*alignlist[i]["identity"]
                    lis[i]["from"] = -1
                    lis[i]["diff"] = 0
    
                    for j in range(i):
                        if lis[j]["used"] or (lis[j]["from"] >=0 and lis[lis[j]["from"]]["align"][highfield] >= lis[i]["align"][lowfield]):
                            continue
                        leni = lis[i]["align"][highfield] - lis[i]["align"][lowfield] + 1
                        lenj = lis[j]["align"][highfield] - lis[j]["align"][lowfield] + 1
                        olap = lis[j]["align"][highfield] - lis[i]["align"][lowfield] + 1
                        if olap < 0:
                            olap = 0
                        diff = lis[j]["diff"]
                        if lis[j]["align"]["targetstart"] < lis[i]["align"]["targetstart"]:
                            diff = diff + lis[j]["align"]["targetend"] - lis[i]["align"]["targetstart"]
                        else:
                            diff = diff + lis[i]["align"]["targetend"] - lis[j]["align"]["targetstart"]
                        if lis[j]["align"]["querylow"] < lis[i]["align"]["querylow"]:
                            diff = diff + lis[j]["align"]["queryhigh"] - lis[i]["align"]["querylow"]
                        else:
                            diff = diff + lis[i]["align"]["queryhigh"] - lis[j]["align"]["querylow"]
    
                        olapfraction = max(olap/leni, olap/lenj)
                        if olapfraction <= maxoverlap:
                            score = lis[j]["score"] + (leni - olap)*alignlist[i]["identity"]*alignlist[i]["identity"]
                        else:
                            score = -1.0
    
                        if score > lis[i]["score"] or (score == lis[i]["score"] and diff < lis[i]["diff"]):
                            lis[i]["from"] = j
                            lis[i]["score"] = score
                            lis[i]["diff"] = diff
                if updatebest(lis, numaligns, allbest):
                    break
    
            numbests = len(allbest)
            eqc = 0
            while eqc < numbests:
                if lis[allbest[eqc]]["diff"] != lis[allbest[0]]["diff"]:
                    break
                eqc = eqc + 1
            # need to have this pick a random number between 0 and eqc instead of using 0
            bestpick = 0
    
            i = allbest[bestpick]
            while i != -1:
                filteredaligns.insert(0, lis[i]["align"])
                i = lis[i]["from"]
        else:
            return -1
        
    return filteredaligns
        #foreach my $rh_entry_pair (@{$ra_entry_pairs}) {
            #my $refentry = $rh_entry_pair->{ref_entry};
            #my $queryentry = $rh_entry_pair->{query_entry};
            #print STDERR "Flagging pair $refentry $queryentry\n";
            #my $ra_aligns = $rh_entry_pair->{aligns};
            #foreach my $rh_align (@{$ra_aligns}) {
                #$rh_align->{query_redundant} = 1;
                #$rh_align->{ref_redundant} = 1;
            #}
    #
            ## find and mark query_redundant:
            #my @sorted_aligns = sort {$a->{query_low} <=> $b->{query_low}} @{$ra_aligns};
    #
            ## mark alignments that are non-redundant by query coord:
            #my @query_starts = map {$_->{query_start}} @{$ra_aligns};
            #my @query_ends = map {$_->{query_end}} @{$ra_aligns};
    #
            ## sorted list of all endpoints:
            #my $no_query_ends = @query_ends;
            #my $no_query_starts = @query_starts;
            #my $expected_total = $no_query_ends + $no_query_starts;
            #my @all_query_ends = @query_starts;
            #push @all_query_ends, @query_ends;
    #
            #@all_query_ends = sort {$a <=> $b} @all_query_ends;
    #
            #my $actual_total = @all_query_ends;
            #print STDERR "Expected total $expected_total actual total $actual_total\n";
    #
            #my @sweep = ();
            #my $last_end = 0;
            #for (my $i=0; $i<=$#all_query_ends; $i++) {
                #my $no_sweep = @sweep;
                #print STDERR "At endpoint $i with $no_sweep in sweep\n";
                #my $this_end = $all_query_ends[$i];
                #next if ($this_end == $last_end); # don't repeat end values
                #$last_end = $this_end; # new end to process
                #while( (@sorted_aligns) && ($sorted_aligns[0]->{query_low}==$this_end )) {
                    #push @sweep, shift @sorted_aligns;
                #}
    #
                #my @temp_sweep = ();
                #my $max_score = 0;
                #foreach my $rh_align (@sweep) {
                    #if ($rh_align->{query_high} != $this_end) {
                        #push @temp_sweep, $rh_align;
                        #if (($rh_align->{query_high} - $rh_align->{query_low}) > $max_score) {
                            #$max_score = $rh_align->{query_high} - $rh_align->{query_low};
                        #}
                    #}
                #}
                #@sweep = @temp_sweep; # keeps only segs that don't end at this end
                #foreach my $rh_align (@sweep) { # mark max score aligns as "good"--scores should have been stored in hash
                    #if (($rh_align->{query_high} - $rh_align->{query_low}) == $max_score) {
                        #$rh_align->{query_redundant} = 0;
                    #}
                #}
            #}
    #
            ## find and mark ref_redundant:
            #@sorted_aligns = sort {$a->{ref_start} <=> $b->{ref_start}} @{$ra_aligns};
    #
            ## mark alignments that are non-redundant by ref coord:
            #my @ref_starts = map {$_->{ref_start}} @{$ra_aligns};
            #my @ref_ends = map {$_->{ref_end}} @{$ra_aligns};
    #
            ## sorted list of all endpoints:
            #my @all_ref_ends = @ref_starts;
            #push @all_ref_ends, @ref_ends;
            #@all_ref_ends = sort {$a <=> $b} @all_ref_ends;
    #
            #@sweep = ();
            #$last_end = 0;
            #for (my $i=0; $i<=$#all_ref_ends; $i++) {
                #my $no_sweep = @sweep;
                #print STDERR "At endpoint $i with $no_sweep in sweep\n";
                #my $this_end = $all_ref_ends[$i];
                #next if ($this_end == $last_end); # don't repeat end values
                #$last_end = $this_end;
                #while( (@sorted_aligns) && ($sorted_aligns[0]->{ref_start}==$this_end )) {
                    #push @sweep, shift @sorted_aligns;
                #}
    #
                #my @temp_sweep = ();
                #my $max_score = 0;
                #foreach my $rh_align (@sweep) {
                    #if ($rh_align->{ref_end} != $this_end) {
                        #push @temp_sweep, $rh_align;
                        #if (($rh_align->{ref_end} - $rh_align->{ref_start}) > $max_score) {
                            #$max_score = $rh_align->{ref_end} - $rh_align->{ref_start};
                        #}
                    #}
                #}
                #@sweep = @temp_sweep; # keeps only segs that don't end at this end
                #foreach my $rh_align (@sweep) { # mark max score aligns as "good"--scores should have been stored in hash
                    #if (($rh_align->{ref_end} - $rh_align->{ref_start}) == $max_score) {
                        #$rh_align->{ref_redundant} = 0;
                    #}
                #}
            #}
        #}
    #}


def create_align_by_entry_dict(alignlist:list, entrytype:str):

    alignbyentrydict = {}
    for align in alignlist:
        entry = align[entrytype]
        if entry not in alignbyentrydict:
            alignbyentrydict[entry] = [align]
        else:
            alignbyentrydict[entry].append(align)

    return alignbyentrydict

def updatebest(lis:list, n:int, allbest:list):
    if n==0:
        return False
    for best in range(n):
        if not lis[best]["used"]:
            break
    for i in range(best+1, n, 1):
        if not lis[i]["used"] and (lis[i]["score"] > lis[best]["score"] or (lis[i]["score"]==lis[best]["score"] and lis[i]["diff"] < lis[best]["diff"]) ):
            best = i
    if len(allbest) > 0 and lis[allbest[0]]["score"] > lis[best]["score"]:
        return False

    allbest.append(best)
    i = best
    while i != -1:
        lis[i]["used"] = True
        i = lis[i]["from"]

    return True

# This routine, ported from mummer's show-diff C++ routine "PrintDiff", steps through reference
# ordered alignments (which have been previously filtered to be RLIS), and analyses the 
# alignment breaks to determine what type they are.
#
# Inputs: The list of alignments passed to the routine should contain a dictionary for each align
# with fields: target, targetstart, targetend, targetlength, query, querystart, queryend, querylength,
# and strand, which should be "F" or "R". All coordinates are assumed to be 1-based.
#
# Return value: The function returns a list with the same alignments passed as input, except the list 
# doesn't contain any of the aligns that are part of the QLIS or RLIS, and it is sorted in order of
# increasing query or reference lower-coordinate.
#

def find_diffs(alignlist:list):
    
    alignbyentrydict = create_align_by_entry_dict(alignlist, entrytype="target")

    inversionends = []
    lisjumps = []
    for target in alignbyentrydict.keys():
        entryaligns = alignbyentrydict[target]
        prevalign = None
        for align in entryaligns:
            # no breaks possible in first align, so just record as prevalign and go to second:
            if not prevalign:
                prevalign = align
                continue
            # break analysis
            gapR = align["targetstart"] - prevalign["targetend"] - 1
            prevquery = prevalign["query"]
            alignquery = align["query"]

            # breaks between aligns to the same scaffold:
            if prevquery == alignquery:
                if prevalign["strand"] != align["strand"]:
                    inversionends.append({"target":target, "query":query, "prevstrand":prevalign["strand"], "alignstrand":align["strand"], "prevtargetend":prevalign["targetend"], "aligntargetstart":align["targetstart"]})
                else:
                    lisjumps.append({"target":target, "query":query, "prevstrand":prevalign["strand"], "alignstrand":align["strand"], "prevtargetend":prevalign["targetend"], "aligntargetstart":align["targetstart"], "prevqueryend":prevalign["queryend"], "alignquerystart":align["querystart"]})





import os
import re
import sys
import random
import pybedtools
import logging
import datetime
from collections import namedtuple
from GQC import seqparse
from GQC import alignparse
from GQC import bedtoolslib
from GQC import coverage

# create namedtuple for bed intervals:
varianttuple = namedtuple('varianttuple', ['chrom', 'start', 'end', 'name', 'vartype', 'excluded', 'qvscore']) 

logger = logging.getLogger(__name__)

def classify_errors(refobj, queryobj, variants, hetsites, outputdict, benchparams, stats, args)->str:

    bencherrorfile = outputdict["bencherrortypebed"]
    stats["singlebasecounts"] = {}
    stats["indellengthcounts"] = {}
    stats["totalerrorsinaligns"] = 0

    testerrorfile = outputdict["testerrortypebed"]
    tfh = open(testerrorfile, "w")

    xfh = None
    if "benchexcludederrortypebed" in outputdict.keys():
        excludederrorfile = outputdict["benchexcludederrortypebed"]
        xfh = open(excludederrorfile, "w")
    if args.vcf:
        bencherrorvcf = bencherrorfile.replace(".bed", "")
        bencherrorvcf = bencherrorvcf + ".vcf"
        vfh = open(bencherrorvcf, "w")
        vcfheader = vcf_header(args)
        vfh.write(vcfheader)
        if "assembly" in args:
            samplename = args.assembly
        elif "qname" in args:
            samplename = args.qname
        else:
            samplename = "SAMPLE"

        vfh.write("#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT\t" + samplename + "\n")

    with open(bencherrorfile, "w") as efh:
        for variant in variants:
            namefields = variant.name.split("_")
            numfields = len(namefields)
            contigname = "_".join(namefields[0:numfields-4])
            pos = int(namefields[-4])
            refallele = namefields[-3]
            altallele = namefields[-2]
            if namefields[-1] == "F":
                alignstrand = '+'
            else:
                alignstrand = '-'
            #logger.debug("Splitting variant name " + variant.name + " and found contigname " + contigname)
            varname = variant.chrom + "_" + str(int(variant.start) + 1) + "_" + refallele + "_" + altallele

            if hetsites and varname in hetsites.keys():
                errortype = 'PHASING'
                errortypecolor = '255,0,0'
            else:
                errortype = 'CONSENSUS'
                errortypecolor = '0,0,255'

            benchvarstart = variant.start
            benchvarend = variant.end
            if refallele == "*":
                benchvarstart = benchvarstart - 1

            varqvscore = "1000"
            if variant.qvscore is not None:
                varqvscore = str(variant.qvscore)

            if xfh and variant.excluded:
                xfh.write(variant.chrom + "\t" + str(benchvarstart) + "\t" + str(benchvarend) + "\t" + varname + "\t" + varqvscore + "\t" + alignstrand + "\t" + str(benchvarstart) + "\t" + str(benchvarend) + "\t" + errortypecolor + "\t" + errortype + "\t" + variant.vartype + "\t" + variant.name + "\n")
                if args.vcf:
                    vcfrecord = vcf_format(variant, refobj, queryobj)
                    vfh.write(vcfrecord)
            else:
                efh.write(variant.chrom + "\t" + str(benchvarstart) + "\t" + str(benchvarend) + "\t" + varname + "\t" + varqvscore + "\t" + alignstrand + "\t" + str(benchvarstart) + "\t" + str(benchvarend) + "\t" + errortypecolor + "\t" + errortype + "\t" + variant.vartype + "\t" + variant.name + "\n")
                if args.vcf:
                    vcfrecord = vcf_format(variant, refobj, queryobj)
                    vfh.write(vcfrecord)

            tfh.write(contigname + "\t" + str(pos-1) + "\t" + str(pos - 1 + len(altallele)) + "\t" + variant.name + "\t" + varqvscore + "\t" + alignstrand + "\t" + str(pos-1) + "\t" + str(pos - 1 + len(altallele)) + "\t" + errortypecolor + "\t" + errortype + "\t" + variant.vartype + "\t" + varname + "\n")
            stats["totalerrorsinaligns"] = stats["totalerrorsinaligns"] + 1
                
            # tally statistics for non-phasing errors:

            if variant.excluded or errortype != "CONSENSUS":
                continue
            if variant.vartype == "SNV":
                snvkey = refallele + "_" + altallele
                if snvkey in stats["singlebasecounts"]:
                    stats["singlebasecounts"][snvkey] = stats["singlebasecounts"][snvkey] + 1
                else:
                    stats["singlebasecounts"][snvkey] = 1
            else:
                trueref = refallele.replace('*', '')
                truealt = altallele.replace('*', '')
                lengthdiff = len(truealt) - len(trueref)
                if lengthdiff in stats["indellengthcounts"]:
                    stats["indellengthcounts"][lengthdiff] = stats["indellengthcounts"][lengthdiff] + 1
                else:
                    stats["indellengthcounts"][lengthdiff] = 1
    tfh.close()
    if xfh:
        xfh.close()

    if args.vcf:
        vfh.close()

    return stats

def vcf_format(variant, refobj, queryobj):

    chrom = variant.chrom
    start = variant.start # zero-based
    refpos = start + 1
    namefields = variant.name.split("_")
    numfields = len(namefields)
    contigname = "_".join(namefields[0:numfields-4])
    contigpos = int(namefields[-4])
    refallele = namefields[-3]
    altallele = namefields[-2]
    if namefields[-1] == "F":
        alignstrand = '+'
    else:
        alignstrand = '-'
    filterfield = 'PASS'
    if variant.excluded:
        filterfield = 'EXCLUDED'

    refallele = refallele.replace("*", "")
    altallele = altallele.replace("*", "")
    while len(refallele) > 0 and len(altallele) > 0 and refallele[-1] == altallele[-1]:
        refallele = refallele[:-1]
        altallele = altallele[:-1]

    if refallele == "" or altallele == "":
        refpos = refpos - 1
        refallele = refobj.fetch(reference=chrom, start=refpos-1, end=refpos) + refallele
        refallele = refallele.upper()
        if alignstrand == '+':
            contigpos = contigpos - 1
            altallele = queryobj.fetch(reference=contigname, start=contigpos-1, end=contigpos) + altallele
            altallele = altallele.upper()
        else:
            if len(refallele) == 1:
                contigend = contigpos + len(altallele)
            else:
                contigend = contigpos + len(altallele) + 1
            altbase = queryobj.fetch(reference=contigname, start=contigend-1, end=contigend)
            altbase = seqparse.revcomp(altbase)
            altbase = altbase.upper()
            altallele = altbase + altallele
            logger.debug("Reverse strand empty allele adjustment: fetch query " + contigname + ":" + str(contigend) + "-" + str(contigend) + " gives " + refallele + "/" +  altallele)
            logger.debug(chrom + "\t" + str(refpos) + "\t" + variant.name + "\t" + refallele + "\t" + altallele + "\t.\t"  + filterfield + "\t.\tGT\t1")

    return chrom + "\t" + str(refpos) + "\t" + variant.name + "\t" + refallele + "\t" + altallele + "\t.\t"  + filterfield + "\t.\tGT\t1\n"

def vcf_header(args):

    if 'benchmark' in args:
        benchname = args.benchmark
    elif 'rname' in args:
        benchname = args.rname
    else:
        benchname = 'REF'
    date = datetime.datetime.now()
    datestring = date.strftime("%Y%m%d")
    header_string = "##fileformat=VCFv4.5\n##fileDate=" + datestring + "\n##source=GQC\n##reference=" + benchname + "\n##FILTER=<ID=EXCLUDED, Description=\"In excluded region of the benchmark reference\">\n##FORMAT=<ID=GT,Number=1,Type=String,Description=\"Genotype\">\n"

    return header_string

def gather_mononuc_stats(coveredmononucbedfile:str, mononucstatsfile:str):

    p = {}
    p["A"] = re.compile("^[aA]+$")
    p["T"] = re.compile("^[tT]+$")
    p["C"] = re.compile("^[cC]+$")
    p["G"] = re.compile("^[gG]+$")
    result = {}
    sfh = open(mononucstatsfile, "w")
    with open(coveredmononucbedfile, "r") as mfh:
        mononucline = mfh.readline()
        while mononucline:
            mononucline = mononucline.rstrip()
            mnfields = mononucline.split("\t")
            if len(mnfields)==12:
                [chrom, start, end, name, score, strand, widestart, wideend, color, chrom_b, start_b, end_b] = mnfields
            elif len(mnfields)==21:
                [chrom, start, end, name, score, strand, widestart, wideend, color, chrom_b, start_b, end_b, variant_name, score_b, strand_b, widestart_b, wideend_b, color_b, error_type, variant_type, queryvariantname] = mnfields
            else:
                logger.critical("Unexpected number of fields (" + str(len(mnfields)) + ") in file " + coveredmononucbedfile + " -- should be 12 or 21")
                print("Unexpected number of fields (" + str(len(mnfields)) + ") in file " + coveredmononucbedfile + " -- should be 12 or 21")
                exit(1)
            runlength = int(end) - int(start)
            namefields = name.split("_")
            repeatedbase = namefields[-1]
            if chrom_b == '.':
                result[name] = {'base':repeatedbase, 'length':runlength, 'assemblylength':runlength, 'type':'CORRECT'}
                sfh.write(name + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(runlength) + "\tCORRECT\n")
            else:
                variantnamefields = variant_name.split("_")
                refbases = variantnamefields[-2]
                altbases = variantnamefields[-1]
                if refbases == "*" and p[repeatedbase].match(altbases): # increased length
                    newlength = runlength + len(altbases)
                    result[name] = {'base':repeatedbase, 'length':runlength, 'assemblylength':newlength, 'type':error_type}
                    sfh.write(name + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(newlength) + "\t" + error_type + "\n")
                elif altbases == "*" and p[repeatedbase].match(refbases): # decreased length
                    newlength = runlength - len(refbases)
                    result[name] = {'base':repeatedbase, 'length':runlength, 'assemblylength':newlength, 'type':error_type}
                    sfh.write(name + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(newlength) + "\t" + error_type + "\n")
                elif p[repeatedbase].match(refbases) and p[repeatedbase].match(altbases): # expanded notation
                    newlength = len(altbases)
                    sfh.write(name + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(newlength) + "\t" + error_type + "\n")
                    result[name] = {'base':repeatedbase, 'length':runlength, 'assemblylength':newlength, 'type':error_type}
                else: # complex error
                    newlength = -1
                    result[name] = {'base':repeatedbase, 'length':runlength, 'assemblylength':newlength, 'type':error_type}
                    sfh.write(name + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(newlength) + "\t" + error_type + "\n")

            mononucline = mfh.readline()

    return result

def assess_mononuc_read_coverage(align_obj, mononucbedfile, outputdict, bedintervals, hetsitedict, args):

    p = {}
    p["A"] = re.compile("^[aA]+$")
    p["T"] = re.compile("^[tT]+$")
    p["C"] = re.compile("^[cC]+$")
    p["G"] = re.compile("^[gG]+$")

    max_reads_per_mononuc = args.maxmononucreads

    mononucstatsfile = outputdict["mononucstatsfile"]

    mononucbedintervals = pybedtools.BedTool(mononucbedfile)
    if bedintervals is not None:
        includedmononucbeds = bedtoolslib.intersectintervals(mononucbedintervals, bedintervals, wa=True)
        includedmononucbeds.saveas(outputdict["includedmononucfile"])
    else:
        mononucbedintervals.saveas(outputdict["includedmononucfile"])

    mononucdict = {}
    msfh = open(mononucstatsfile, "w", buffering=1)
    # for each mononuc run in the benchmark, retrieve reads, count length of mononuc, classify as CORRECT, HET, ERROR, or COMPLEX
    with open(outputdict["includedmononucfile"], "r") as mfh:
        mononucline = mfh.readline()
        while mononucline:
            mononucline = mononucline.rstrip()
            [chrom, start, end, name, score, strand, widestart, wideend, color] = mononucline.split("\t")
            runlength = int(end) - int(start)
            if runlength not in mononucdict:
                mononucdict[runlength] = {}
            namefields = name.split("_")
            repeatedbase = namefields[-1]
            refalleleseq = repeatedbase * runlength
            readsforthismononuc = 0
            for readalign in align_obj.fetch(contig=chrom, start=int(start), stop=int(end)):
                if max_reads_per_mononuc > 0 and readsforthismononuc >= max_reads_per_mononuc:
                    break
                if args.downsample is not None and random.random() >= args.downsample:
                    continue
                queryseq = readalign.query_sequence
                if readalign.is_secondary or queryseq is None:
                    continue
                readname = readalign.query_name
                if readalign.reference_start >= int(start) or readalign.reference_end <= int(end):
                    continue
                pairs = readalign.get_aligned_pairs()
                if len(pairs) == 0:
                    continue
                # find zero-based read pos of base aligned to ref base one before mononuc:
                readstart = find_readpos_in_pairs(pairs, int(start)-1)
                # find zero-based read pos of base aligned to ref base one after mononuc:
                readend = find_readpos_in_pairs(pairs, int(end))
                if readstart is not None and readend is not None:
                    queryseq = queryseq.upper()
                    alleleseq = queryseq[readstart+1:readend]
                    # extend the read sequence so 5' and 3' to include all of repeated bases
                    #if readstart >=0:
                        #logger.debug("Checking " + queryseq[readstart] + " vs " + repeatedbase)
                    while readstart >= 0 and queryseq[readstart] == repeatedbase:
                        #logger.debug("Extending " + alleleseq + " one base")
                        alleleseq = repeatedbase + alleleseq
                        readstart = readstart - 1
                    #if readend < len(queryseq):
                        #logger.debug("Checking " + queryseq[readend] + " vs " + repeatedbase)
                    while readend < len(queryseq) and queryseq[readend] == repeatedbase:
                        #logger.debug("Extending " + alleleseq + " one base")
                        alleleseq = alleleseq + repeatedbase
                        readend = readend + 1

                    if p[repeatedbase].match(alleleseq):
                        numbases = len(alleleseq)
                        matchtype = "CORRECT"
                        if numbases != runlength:
                            potentialhetname = chrom + "_" + str(int(start)+1) + "_" + refalleleseq + "_" + alleleseq
                            if potentialhetname in hetsitedict:
                                matchtype = "HET"
                            else:
                                matchtype = "ERROR"
                    else:
                        numbases = -1
                        matchtype = "COMPLEX"

                    msfh.write(chrom + "\t" + str(start) + "\t" + str(end) + "\t" + readname + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(numbases) + "\t" + matchtype + "\n")
                    if numbases not in mononucdict[runlength]:
                        mononucdict[runlength][numbases] = {}
                    if matchtype not in mononucdict[runlength][numbases]:
                        mononucdict[runlength][numbases][matchtype] = 1
                    else:
                        mononucdict[runlength][numbases][matchtype] = mononucdict[runlength][numbases][matchtype] + 1
                    readsforthismononuc = readsforthismononuc + 1
                else:
                    logger.debug(readname + " is unaligned at one endpoint of interval " + str(start) + " to " + str(end) + "!")
                    if readstart is None:
                        logger.warning(readname + " " + str(readalign.reference_start) + "-" + str(readalign.reference_end) + " does not have a start")
                    if readend is None:
                        logger.warning(readname + " " + str(readalign.reference_start) + "-" + str(readalign.reference_end) + " does not have a end")

            mononucline = mfh.readline()

    return mononucdict

def assess_str_read_coverage_require_flank(strlabel, align_obj, refobj, strbedfile, outputdict, bedintervals, hetsitedict, args):

    max_reads_per_str = args.maxstrreads

    includedstrfile = outputdict["includedstrprefix"] + "." + strlabel + ".included.bed"
    strstatsfile = outputdict["strstatsprefix"] + "." + strlabel + ".stats.out"

    logger.info("Writing " + strlabel + " STR data to " + strstatsfile)

    # if the file already exists, don't recreate it
    if os.path.exists(strstatsfile) and os.path.getsize(strstatsfile) > 0:
        strdict = retrieve_stats_from_strstats_file(strstatsfile)
        return strdict

    strbedintervals = pybedtools.BedTool(strbedfile)
    if bedintervals is not None:
        includedstrbeds = bedtoolslib.intersectintervals(strbedintervals, bedintervals, wa=True)
        includedstrbeds.saveas(includedstrfile)
    else:
        strbedintervals.saveas(includedstrfile)

    print("Saved included STR bed file. Will now open statsfile")
    strdict = {}
    sfh = open(strstatsfile, "w", buffering=1)
    # for each str run in the str bedfile, retrieve reads, count length of str, classify as CORRECT, HET, ERROR, or COMPLEX
    with open(includedstrfile, "r") as ifh:
        strline = ifh.readline()
        while strline:
            strline = strline.rstrip()
            strfields = strline.split("\t")
            chrom = strfields[0]
            start = int(strfields[1])
            end = int(strfields[2])
            name = strfields[3]

            # need to extract benchmark's flanking bases here for comparison later!!!
            benchleftflankseq = refobj.fetch(reference=chrom, start=start-5, end=start)
            benchleftflankseq = benchleftflankseq.upper()
            benchrightflankseq = refobj.fetch(reference=chrom, start=end, end=end+5)
            benchrightflankseq = benchrightflankseq.upper()
            runlength = end - start
            if runlength not in strdict:
                strdict[runlength] = {}
            namefields = name.split("_")
            repeatedbases = namefields[-1]
            refalleleseq = refobj.fetch(reference=chrom, start=start, end=end)
            readsforthisstr = 0
            # for each read, attempt to find read positions aligned to the bases immediately flanking
            # the STR. If one side and/or the other has no aligning base (due either to a deletion or
            # to clipped sequence) then check for clipping beginning at that base, and
            # extend the STR-aligned read sequence with clipped sequence to include adjacent 
            # bases that extend the repeat plus 5 bp additional (this is NIST's so-called "slop" length).
            # If a flanking base *does* have a read base aligned to it, also extend that end to 
            # include repeated bases plut 5 bp slop.
            for readalign in align_obj.fetch(contig=chrom, start=start, stop=end):
                if max_reads_per_str > 0 and readsforthisstr >= max_reads_per_str:
                    break
                if args.downsample is not None and random.random() >= args.downsample:
                    continue
                # query_sequence includes soft clipped bases:
                queryseq = readalign.query_sequence
                if readalign.is_secondary or queryseq is None:
                    continue
                queryseq = queryseq.upper()
                readname = readalign.query_name
                # if the read doesn't align to the entire STR, skip it:
                if readalign.reference_start is None or readalign.reference_end is None:
                    logger.debug("Read " + readname + " has no reference_start or reference_end for " + chrom + ":" + str(start) + "-" + str(end))
                    continue
                if readalign.reference_start > start or readalign.reference_end < end:
                    continue
                # if the read doesn't align to the flanking bases, it must be clipped:
                cigartuples = readalign.cigartuples
                if ((readalign.reference_start == start and cigartuples[0][0] != 4) or
                    (readalign.reference_end == end and cigartuples[len(cigartuples)-1][0] != 4)):
                    logger.debug("Read " + readname + " aligns just to end of " + name + " but is not soft clipped--skipping")
                    continue
                pairs = readalign.get_aligned_pairs()
                if len(pairs) == 0:
                    continue
                # find zero-based read pos of base aligned to ref base one before STR:
                if readalign.reference_start < start:
                    readstart = find_readpos_in_pairs(pairs, start-1)
                elif readalign.query_alignment_start > 0: # must be clipped--set start to one efore base aligning to reference start:
                    readstart = readalign.query_alignment_start - 1
                # find zero-based read pos of base aligned to ref base one after STR:
                if readalign.reference_end > end:
                    readend = find_readpos_in_pairs(pairs, end)
                elif readalign.query_alignment_end < len(queryseq):
                    readend = readalign.query_alignment_end + 1

                if readstart is not None and readend is not None:
                    # extend the read sequence to 5' and 3' so long as it matches the STR. "correctcomp" is True or False depending
                    # on whether the alleleseq is a faithful repeat (even if flanking copies are partial) of the repeatedbases passed to it:
                    [alleleseq, readstart, readend, correctcomp] = widen_str_on_readseq(queryseq, readstart, readend, repeatedbases)

                    # check that read has 5 flanking base pairs on either side of the HP:
                    if readstart < 5 or readend > len(queryseq)-5:
                        logger.debug("Skipping align of read " + readname + " which has fewer than 5 bp outside of HP repeat for " + chrom + ":" + str(start) + "-" + str(end))
                        continue
                    else:
                        leftflankseq = queryseq[readstart-4:readstart+1]
                        leftflankseq = leftflankseq.upper()
                        rightflankseq = queryseq[readend:readend+5]
                        rightflankseq = rightflankseq.upper()
                    matchtype = "UNKNOWN"
                    if correctcomp:
                        numbases = len(alleleseq)
                        if numbases != runlength:
                            potentialhetname = chrom + "_" + str(start+1) + "_" + refalleleseq + "_" + alleleseq
                            if potentialhetname in hetsitedict:
                                matchtype = "HET"
                            else:
                                matchtype = "LENGTHERROR"
                                logger.debug("Potential het " + potentialhetname + " is not a het")
                        else:
                            if leftflankseq == benchleftflankseq and rightflankseq == benchrightflankseq:
                                matchtype = "CORRECT"
                            else:
                                logger.debug(leftflankseq + "/" + rightflankseq + " doesnt match " + benchleftflankseq + "/" + benchrightflankseq)
                                matchtype = "FLANKERROR"
                    else: # discrepancies within the HP 
                        numbases = -1
                        matchtype = "COMPLEX"

                    sfh.write(chrom + "\t" + str(start) + "\t" + str(end) + "\t" + readname + "\t" + repeatedbases + "\t" + str(runlength) + "\t" + str(numbases) + "\t" + matchtype + "\n")
                    if numbases not in strdict[runlength]:
                        strdict[runlength][numbases] = {}
                    if matchtype not in strdict[runlength][numbases]:
                        strdict[runlength][numbases][matchtype] = 1
                    else:
                        strdict[runlength][numbases][matchtype] = strdict[runlength][numbases][matchtype] + 1
                    readsforthisstr = readsforthisstr + 1
                else:
                    logger.debug(readname + " is unaligned at one endpoint of interval " + str(start) + " to " + str(end) + " and has fewer than 5 clipped bases!")
                    if readstart is None:
                        logger.warning(readname + " " + str(readalign.reference_start) + "-" + str(readalign.reference_end) + " does not have a start")
                    if readend is None:
                        logger.warning(readname + " " + str(readalign.reference_start) + "-" + str(readalign.reference_end) + " does not have a end")

            strline = ifh.readline()

    return strdict

def assess_mononuc_read_coverage_require_flank(align_obj, refobj, mononucbedfile, outputdict, bedintervals, hetsitedict, args):

    p = {}
    p["A"] = re.compile("^[aA]+$")
    p["T"] = re.compile("^[tT]+$")
    p["C"] = re.compile("^[cC]+$")
    p["G"] = re.compile("^[gG]+$")

    max_reads_per_mononuc = args.maxmononucreads

    mononucstatsfile = outputdict["mononucstatsfile"]

    # if the file already exists, don't recreate it
    if os.path.exists(mononucstatsfile):
        mononucdict = retrieve_stats_from_strstats_file(mononucstatsfile)
        return mononucdict

    mononucbedintervals = pybedtools.BedTool(mononucbedfile)
    if bedintervals is not None:
        includedmononucbeds = bedtoolslib.intersectintervals(mononucbedintervals, bedintervals, wa=True)
        includedmononucbeds.saveas(outputdict["includedmononucfile"])
    else:
        mononucbedintervals.saveas(outputdict["includedmononucfile"])

    mononucdict = {}
    msfh = open(mononucstatsfile, "w", buffering=1)
    # for each mononuc run in the benchmark, retrieve reads, count length of mononuc, classify as CORRECT, HET, ERROR, or COMPLEX
    with open(outputdict["includedmononucfile"], "r") as mfh:
        mononucline = mfh.readline()
        while mononucline:
            mononucline = mononucline.rstrip()
            [chrom, start, end, name, score, strand, widestart, wideend, color] = mononucline.split("\t")
            # need to extract benchmark's flanking bases here for comparison later!!!
            benchleftflankseq = refobj.fetch(reference=chrom, start=int(start)-5, end=int(start))
            benchleftflankseq = benchleftflankseq.upper()
            benchrightflankseq = refobj.fetch(reference=chrom, start=int(end), end=int(end)+5)
            benchrightflankseq = benchrightflankseq.upper()
            runlength = int(end) - int(start)
            if runlength not in mononucdict:
                mononucdict[runlength] = {}
            namefields = name.split("_")
            repeatedbase = namefields[-1]
            refalleleseq = repeatedbase * runlength
            readsforthismononuc = 0
            # for each read, attempt to find read positions aligned to the bases immediately flanking
            # the mononucleotide. If one side and/or the other has no aligning base (due either to a 
            # deletion or to clipped sequence) then check for clipping beginning at that base, and
            # extend the mononuc-aligned read sequence with clipped sequence to include adjacent 
            # repeated bases plus 5 bp additional (this is NIST's so-called "slop" length).
            # If a flanking base *does* have a read base aligned to it, also extend that end to 
            # include repeated bases plut 5 bp slop.
            for readalign in align_obj.fetch(contig=chrom, start=int(start), stop=int(end)):
                if max_reads_per_mononuc > 0 and readsforthismononuc >= max_reads_per_mononuc:
                    break
                if args.downsample is not None and random.random() >= args.downsample:
                    continue
                # query_sequence includes soft clipped bases:
                queryseq = readalign.query_sequence
                queryseq = queryseq.upper()
                if readalign.is_secondary or queryseq is None:
                    continue
                readname = readalign.query_name
                # if the read doesn't align to the entire mononucleotide, skip it:
                if readalign.reference_start is None or readalign.reference_end is None:
                    logger.debug("Read " + readname + " has no reference_start or reference_end for " + chrom + ":" + str(start) + "-" + str(end))
                    continue
                if readalign.reference_start > int(start) or readalign.reference_end < int(end):
                    continue
                # if the read doesn't align to the flanking bases, it must be clipped:
                cigartuples = readalign.cigartuples
                if ((readalign.reference_start == int(start) and cigartuples[0][0] != 4) or
                    (readalign.reference_end == int(end) and cigartuples[len(cigartuples)-1][0] != 4)):
                    logger.debug("Read " + readname + " aligns just to end of " + name + " but is not soft clipped--skipping")
                    continue
                pairs = readalign.get_aligned_pairs()
                if len(pairs) == 0:
                    continue
                # find zero-based read pos of base aligned to ref base one before mononuc:
                if readalign.reference_start < int(start):
                    readstart = find_readpos_in_pairs(pairs, int(start)-1)
                elif readalign.query_alignment_start > 0: # must be clipped--set start to one base before reference start:
                    readstart = readalign.query_alignment_start - 1
                # find zero-based read pos of base aligned to ref base one after mononuc:
                if readalign.reference_end > int(end):
                    readend = find_readpos_in_pairs(pairs, int(end))
                elif readalign.query_alignment_end < len(queryseq):
                    readend = readalign.query_alignment_end + 1

                if readstart is not None and readend is not None:
                    alleleseq = queryseq[readstart+1:readend]
                    # extend the read sequence so 5' and 3' to include all of repeated bases
                    #if readstart >=0:
                        #logger.debug("Checking " + queryseq[readstart] + " vs " + repeatedbase)
                    while readstart >= 0 and queryseq[readstart] == repeatedbase:
                        #logger.debug("Extending " + alleleseq + " one base")
                        alleleseq = repeatedbase + alleleseq
                        readstart = readstart - 1
                    #if readend < len(queryseq):
                        #logger.debug("Checking " + queryseq[readend] + " vs " + repeatedbase)
                    while readend < len(queryseq) and queryseq[readend] == repeatedbase:
                        #logger.debug("Extending " + alleleseq + " one base")
                        alleleseq = alleleseq + repeatedbase
                        readend = readend + 1

                    # check that read has 5 flanking base pairs on either side of the HP:
                    if readstart < 5 or readend > len(queryseq)-5:
                        logger.debug("Skipping align of read " + readname + " which has fewer than 5 bp outside of HP repeat for " + chrom + ":" + str(start) + "-" + str(end))
                        continue
                    else:
                        leftflankseq = queryseq[readstart-4:readstart+1]
                        leftflankseq = leftflankseq.upper()
                        rightflankseq = queryseq[readend:readend+5]
                        rightflankseq = rightflankseq.upper()
                    matchtype = "UNKNOWN"
                    if p[repeatedbase].match(alleleseq):
                        numbases = len(alleleseq)
                        if numbases != runlength:
                            potentialhetname = chrom + "_" + str(int(start)+1) + "_" + refalleleseq + "_" + alleleseq
                            if potentialhetname in hetsitedict:
                                matchtype = "HET"
                            else:
                                matchtype = "LENGTHERROR"
                        else:
                            if leftflankseq == benchleftflankseq and rightflankseq == benchrightflankseq:
                                matchtype = "CORRECT"
                            else:
                                logger.debug(leftflankseq + "/" + rightflankseq + " doesnt match " + benchleftflankseq + "/" + benchrightflankseq)
                                matchtype = "FLANKERROR"
                    else: # discrepancies within the HP 
                        numbases = -1
                        matchtype = "COMPLEX"

                    msfh.write(chrom + "\t" + str(start) + "\t" + str(end) + "\t" + readname + "\t" + repeatedbase + "\t" + str(runlength) + "\t" + str(numbases) + "\t" + matchtype + "\n")
                    if numbases not in mononucdict[runlength]:
                        mononucdict[runlength][numbases] = {}
                    if matchtype not in mononucdict[runlength][numbases]:
                        mononucdict[runlength][numbases][matchtype] = 1
                    else:
                        mononucdict[runlength][numbases][matchtype] = mononucdict[runlength][numbases][matchtype] + 1
                    readsforthismononuc = readsforthismononuc + 1
                else:
                    logger.debug(readname + " is unaligned at one endpoint of interval " + str(start) + " to " + str(end) + " and has fewer than 5 clipped bases!")
                    if readstart is None:
                        logger.warning(readname + " " + str(readalign.reference_start) + "-" + str(readalign.reference_end) + " does not have a start")
                    if readend is None:
                        logger.warning(readname + " " + str(readalign.reference_start) + "-" + str(readalign.reference_end) + " does not have a end")

            mononucline = mfh.readline()

    return mononucdict

def retrieve_stats_from_strstats_file(filename)->dict:
    strdict = {}

    with open(filename, "r") as sfh:
        statline = sfh.readline()
        while statline:
            statline = statline.rstrip()
            [chrom, start, end, readname, repeatedbases, runlength, numbases, matchtype] = statline.split("\t")
            start = int(start)
            end = int(end)
            runlength = int(runlength)
            if runlength not in strdict:
                strdict[runlength] = {}
            numbases = int(numbases)
            if numbases not in strdict[runlength]:
                strdict[runlength][numbases] = {}
            if matchtype not in strdict[runlength][numbases]:
                strdict[runlength][numbases][matchtype] = 1
            else:
                strdict[runlength][numbases][matchtype] = strdict[runlength][numbases][matchtype] + 1
            statline = sfh.readline()

    return strdict

# "pairs" is a structure created by pysam for an alignment, and contains a list
#   of tuples "consisting of the 0-based offset from the start of the read 
#   sequence followed by the 0-based reference position
#
# This routine searches the list of tuples for the one corresponding to the 
#   desired reference position ("pos"), and returns the read position
#   for that tuple, assuming it exists
#
# Why would a tuple have a ref position of "None"? These tuples are insertions
#   in the read sequence (and deletions from the reference have read pos of 
#   None). In the case of deletions from the reference, the "ifnone" argument
#   can tell the routine whether to report the read position that is one to 
#   the left of the deleted ref location ("lower") or one to the right ("higher")
#
# Note: this routine takes a *zero-based* position as its "pos" argument!
#
def find_readpos_in_pairs(pairs, pos, ifnone="lower"):
    # number of tuples:
    alignlength = len(pairs)
    ilow = 0
    ihigh = alignlength - 1

    hiref = pairs[ihigh][1]
    while hiref is None and ihigh > 0:
        ihigh = ihigh - 1
        hiref = pairs[ihigh][1]
    lowref = pairs[ilow][1]
    while lowref is None and ilow < len(pairs):
        ilow = ilow + 1
        lowref = pairs[ilow][1]

    # these shouldn't really happen
    if lowref is None or hiref is None or lowref > pos or hiref < pos or lowref > hiref:
        return None

    # need to be careful here to avoid an infinite loop:
    lastimid = None
    while ihigh - 1 > ilow and hiref >= pos and lowref <= pos:
        imid = int((ilow + ihigh)/2)
        midref = pairs[imid][1]
        while midref is None:
            if ifnone == "lower":
                imid = imid - 1
            else:
                imid = imid + 1
            midref = pairs[imid][1]
            if (ifnone == "lower" and imid <= ilow) or (ifnone=="higher" and imid >= ihigh):
                break
        if midref is None:
            return None
        if lastimid is not None and lastimid == imid:
            for i in range(ilow, ihigh):
                if pairs[i][1] is not None and pairs[i][1] == pos:
                    return pairs[i][0]
            return None
        
        if midref == pos:
            return pairs[imid][0] # might be None if nothing is aligned here
        elif midref > pos:
            ihigh = imid
            hiref = pairs[ihigh][1]
        elif midref < pos:
            ilow = imid
            lowref = pairs[ilow][1]
        if (pairs[ilow][1] is not None and pairs[ilow][1] > pos) or (pairs[ihigh][1] is not None and pairs[ihigh][1] < pos):
            return None
        lastimid = imid

    return None

def widen_str_on_readseq(readseq:str, readstart:int, readend:int, repbases:str)->list:
    alleleseq = readseq[readstart+1:readend]

    replength = len(repbases) 
    readlength = len(readseq) 
    perfectrepeat = True

    initialalleleseq = alleleseq
    initialreadstart = readstart
    initialreadend = readend

    # check whether read's current allele sequence is purely composed of the repeat:
    alleleindex = 0
    while alleleindex < len(alleleseq):
        allelebase = alleleseq[alleleindex]
        repbaseindex = alleleindex % replength
        repbase = repbases[repbaseindex]
        if repbase != allelebase:
            logger.debug("Allele sequence " + alleleseq + " is not a perfect repeat of " + repbases)
            perfectrepeat = False
            return [alleleseq, readstart, readend, perfectrepeat]
        alleleindex = alleleindex + 1

    # extend to left:
    readindex = readstart
    repbaseindex = (readindex - 1 - readstart) % replength
    while readindex > 0 and readseq[readindex] == repbases[repbaseindex]:
        readstart = readstart - 1
        alleleseq = readseq[readindex] + alleleseq
        readindex = readindex - 1
        repbaseindex = (readindex - 1 - initialreadstart) % replength

    # extend to right:
    readindex = readend
    repbaseindex = (readindex - readstart - 1) % replength
    while readindex < readlength and readseq[readindex] == repbases[repbaseindex]:
        readend = readend + 1
        alleleseq = alleleseq + readseq[readindex]
        readindex = readindex + 1
        repbaseindex = (readindex - readstart - 1) % replength

    if initialalleleseq != alleleseq:
        logger.debug("Extended " + initialalleleseq + " to " + alleleseq + " " + str(initialreadstart)  + "-" + str(initialreadend) + "/" + str(readstart) + "-" + str(readend) )

    return [alleleseq, readstart, readend, perfectrepeat]

def assess_read_align_errors(align_obj, refobj, readerrorfile:str, bedintervals, hetsitedict, args):
   
    if not args.errorfile and not args.rerun:
        refh = open(readerrorfile, "w")
    elif args.errorfile:
        readerrorfile = args.errorfile 

    stats = {}
    stats["totalalignedbases"] = 0
    stats["totalclippedbases"] = 0
    stats["totalerrorsinaligns"] = 0
    stats["singlebasecounts"] = {}
    stats["indellengthcounts"] = {}
    stats["alignedqualscorecounts"] = []
    # position counts are zero-based, i.e., one less than the ordinal position in the read:
    stats["positionsnvcounts"] = []
    stats["positionindelcounts"] = []
    stats["positiontotalcounts"] = []
    stats["snverrorqualscorecounts"] = []
    stats["indelerrorqualscorecounts"] = []
    alignsprocessed = 0

    # are we dealing with just a set of regions? or the whole genome?
    if bedintervals is None:
        bedintervals = [None]

    for benchinterval in bedintervals:
        if benchinterval is not None:
            regionstring = benchinterval.chrom + ":" + str(benchinterval.start + 1) + "-" + str(benchinterval.end)
        else:
            regionstring = None

        for align in align_obj.fetch(region=regionstring):

            if align.is_secondary or align.cigartuples is None or (args.downsample is not None and random.random() >= args.downsample):
                continue

            stats["totalalignedbases"] = stats["totalalignedbases"] + align.reference_length
            if benchinterval is not None: # shorten aligned length contribution if necessary
                if benchinterval.start > align.reference_start:
                    stats["totalalignedbases"] = stats["totalalignedbases"] + align.reference_start - benchinterval.start
                if benchinterval.end < align.reference_end:
                    stats["totalalignedbases"] = stats["totalalignedbases"] - align.reference_end + benchinterval.end

            # TODO: clipping calc should be amended to not count if alignment endpoints are outside desired benchinterval!
            cigartuples = align.cigartuples
            if cigartuples[0][0] in [4, 5]:
                if benchinterval is None or benchinterval.start < align.reference_start:
                    stats["totalclippedbases"] = stats["totalclippedbases"] + cigartuples[0][1]
            if cigartuples[-1][0] in [4, 5]:
                if benchinterval is None or benchinterval.end > align.reference_end:
                    stats["totalclippedbases"] = stats["totalclippedbases"] + cigartuples[-1][1]
            
            if not args.errorfile and not args.rerun:
                query, querystart, queryend, ref, refstart, refend, strand = alignparse.retrieve_align_data(align)
                if strand == "F":
                    queryleft = querystart
                    queryright = queryend
                else:
                    queryleft = queryend
                    queryright = querystart

                # make sure position arrays (zero-base indexed) are long enough to store position counts:
                while len(stats["positionindelcounts"]) < queryend:
                    stats["positionindelcounts"].append(0)
                while len(stats["positionsnvcounts"]) < queryend:
                    stats["positionsnvcounts"].append({})
                while len(stats["positiontotalcounts"]) < queryend:
                    stats["positiontotalcounts"].append(0)
                for position in range(querystart-1, queryend):
                    stats["positiontotalcounts"][position-1] = stats["positiontotalcounts"][position-1] + 1

                hetsites = {}
                hetsitealleles = {} # no need to track het site alleles in this context
                queryobj = None
    
                read_variants = alignparse.align_variants(align, queryobj, query, querystart, queryend, refobj, ref, refstart, refend, strand, hetsites, hetsitealleles, stats["alignedqualscorecounts"], stats["snverrorqualscorecounts"], stats["indelerrorqualscorecounts"], True)
    
                for variant in read_variants:
                    namefields = variant.name.split("_")
                    # this is the read position (from beginning of read, regardless of strand):
                    pos = int(namefields[-4])
                    refallele = namefields[-3]
                    altallele = namefields[-2]
                    if benchinterval is not None and (variant.start < benchinterval.start or variant.end > benchinterval.end):
                        continue
                    if namefields[-1] == "F":
                        alignstrand = '+'
                    else:
                        alignstrand = '-'
                    varname = variant.chrom + "_" + str(int(variant.start) + 1) + "_" + refallele + "_" + altallele
                    varqvscore = "1000"
                    if variant.qvscore is not None:
                        varqvscore = str(variant.qvscore)

        
                    if varname in hetsitedict.keys():
                        errortype = 'HET'
                        errortypecolor = '255,0,0'
                    else:
                        errortype = 'ERROR'
                        errortypecolor = '0,0,255'
    
                    # record error in stats:
                    if errortype == 'ERROR':
                        stats["totalerrorsinaligns"] = stats["totalerrorsinaligns"] + 1
                        if refallele != "*" and altallele != "*" and len(refallele) == 1 and len(altallele) == 1:
                            if alignstrand == "+":
                                readrefallele = refallele
                                readaltallele = altallele
                            else:
                                readrefallele = seqparse.revcomp(refallele)
                                readaltallele = seqparse.revcomp(altallele)

                            snvkey = readrefallele + "_" + readaltallele
                            if snvkey in stats["singlebasecounts"]:
                                stats["singlebasecounts"][snvkey] = stats["singlebasecounts"][snvkey] + 1
                            else:
                                stats["singlebasecounts"][snvkey] = 1
                            # add SNV error to tally at this read position:
                            if snvkey in stats["positionsnvcounts"][pos - 1]:
                                stats["positionsnvcounts"][pos-1][snvkey] = stats["positionsnvcounts"][pos-1][snvkey] + 1
                            else:
                                stats["positionsnvcounts"][pos-1][snvkey] = 1
                        else:
                            trueref = refallele.replace('*', '')
                            truealt = altallele.replace('*', '')
                            lengthdiff = len(truealt) - len(trueref)
                            if lengthdiff in stats["indellengthcounts"]:
                                stats["indellengthcounts"][lengthdiff] = stats["indellengthcounts"][lengthdiff] + 1
                            else:
                                stats["indellengthcounts"][lengthdiff] = 1
                            #stats["positionindelcounts"][pos-1] = stats["positionindelcounts][pos-1] + 1
    
                    refh.write(variant.chrom + "\t" + str(variant.start) + "\t" + str(variant.end) + "\t" + varname + "\t" + varqvscore + "\t" + alignstrand + "\t" + str(variant.start) + "\t" + str(variant.end) + "\t" + errortypecolor + "\t" + errortype + "\t" + variant.name + "\n")
    
            alignsprocessed = alignsprocessed + 1
            if alignsprocessed == 100000*int(alignsprocessed/100000):
                logger.debug("Processed " + str(alignsprocessed) + " aligns")


    if args.errorfile or args.rerun:
        with open(readerrorfile, "r") as efh:
            errorline = efh.readline()
            while errorline:
                errorline = errorline.rstrip()
                [chrom, start, end, varname, score, strand, widestart, wideend, color, variant_type, queryvariantname] = errorline.split("\t")
                if variant_type == "HET":
                    errorline = efh.readline()
                    continue

                namefields = varname.split("_")
                refallele = namefields[-2]
                altallele = namefields[-1]
                
                # record error in stats:
                stats["totalerrorsinaligns"] = stats["totalerrorsinaligns"] + 1
                if refallele != "*" and altallele != "*" and len(refallele) == 1 and len(altallele) == 1:
                    snvkey = refallele + "_" + altallele
                    if snvkey in stats["singlebasecounts"]:
                        stats["singlebasecounts"][snvkey] = stats["singlebasecounts"][snvkey] + 1
                    else:
                        stats["singlebasecounts"][snvkey] = 1
                else:
                    trueref = refallele.replace('*', '')
                    truealt = altallele.replace('*', '')
                    lengthdiff = len(truealt) - len(trueref)
                    if lengthdiff in stats["indellengthcounts"]:
                        stats["indellengthcounts"][lengthdiff] = stats["indellengthcounts"][lengthdiff] + 1
                    else:
                        stats["indellengthcounts"][lengthdiff] = 1
                errorline = efh.readline()

    return stats

def write_combined_ref_error_files(outputdir:str, comparisondata:dict, comparisonoutputfiles:dict, refbedobj, args):

    logger.info("Writing error bed files")
    combinedreferrorfile = outputdir + "/" + args.qname + "_vs_" + args.rname + ".refdiscrepancies.sort.bed"
    nolongstrcombreferrorfile = outputdir + "/" + args.qname + "_vs_" + args.rname + ".refdiscrepancies.nolongstrs.sort.bed"
    nostrsubq30combreferrorfile = outputdir + "/" + args.qname + "_vs_" + args.rname + ".refdiscrepancies.nolongstrs.subq30.sort.bed"
    nostrsubq40combreferrorfile = outputdir + "/" + args.qname + "_vs_" + args.rname + ".refdiscrepancies.nolongstrs.subq40.sort.bed"
    nostrsubq50combreferrorfile = outputdir + "/" + args.qname + "_vs_" + args.rname + ".refdiscrepancies.nolongstrs.subq50.sort.bed"

    # write discrepancies in alignments for entire ref genome to file (if necessary)
    if not os.path.exists(combinedreferrorfile):
        combinedreferrorbedstring = ""
        for comparison in comparisondata.keys():
            with open(comparisonoutputfiles[comparison]['referrorbed']) as referrorfh:
                logger.info("Including ref discrepancy file" + comparisonoutputfiles[comparison]['referrorbed'])
                combinedreferrorbedstring = combinedreferrorbedstring + referrorfh.read()
        combinedreferrorobj = pybedtools.BedTool(combinedreferrorbedstring, from_string=True)
        logger.info("Saving to " + combinedreferrorfile)
        combinedreferrorobj.sort().saveas(combinedreferrorfile)
    else:
        combinedreferrorobj = pybedtools.BedTool(combinedreferrorfile)

    # write non-long STR discrepancies in alignments for entire ref genome to file (if necessary)
    if not os.path.exists(nolongstrcombreferrorfile):
        nolongstrints = []
        for errorint in combinedreferrorobj:
            errorname = errorint.name
            erroralleles = errorname.split("_")
            [ref, alt] = erroralleles[-2:]
            if len(ref) < 10 or len(alt) < 10 or abs(len(ref) - len(alt)) > 4:
                nolongstrints.append(errorint)
        nolongstrcombreferrorobj = pybedtools.BedTool(nolongstrints)
        nolongstrcombreferrorobj.saveas(nolongstrcombreferrorfile)
    else:
        nolongstrcombreferrorobj = pybedtools.BedTool(nolongstrcombreferrorfile)

    # write windowed bedfiles with windows having higher levels of non-long STR discrepancies between assemblies:
    # these files will be subtracted from the genome, along with N regions and regions uncovered by alignments, to obtain 
    # regions in the reference covered by high-accuracy aligned query:
    if not os.path.exists(nostrsubq50combreferrorfile):
    
        # find regions with higher levels of non-long STR discrepancies between the assemblies:
        logger.info("Using bin size 100000")
        binsize = 100000

        # can optionally pass a filename to binintervals as last argument to write file
        binnedcoveredintervals = bedtoolslib.binintervals(refbedobj, binsize)
         
        
        #[errorbins, errorbincounts, errorbinindex] = coverage.initiate_bins(refbedobj, binsize, args, includepartial=True)

        #for error in nolongstrcombreferrorobj:
            #chrom = error.chrom
            #start = error.start
            #binstart = binsize*int(start/binsize)
            #binkey = chrom + ":" + str(binstart)
            #if binkey in errorbinindex.keys():
                #errorbincounts[errorbinindex[binkey]] = errorbincounts[errorbinindex[binkey]] + 1
            #else:
                #logger.debug("No bin key for position " + chrom + ":" + str(start))
#
        #errorcount30string = ""
        #errorcount40string = ""
        #errorcount50string = ""
        #for errorbin in errorbins:
            #chrom = errorbin.chrom
            #start = errorbin.start
            #end = errorbin.end
            #binkey = chrom + ":" + str(start)
            #binnumber = errorbinindex[binkey]
            #errorcount = errorbincounts[binnumber]
#
            #if errorcount/(end - start) >= 0.001:
                #errorcount30string = errorcount30string + "\n" + chrom + "\t" + str(start) + "\t" + str(end) + "\t" + str(errorcount)
            #if errorcount/(end - start) >= 0.0001:
                #errorcount40string = errorcount40string + "\n" + chrom + "\t" + str(start) + "\t" + str(end) + "\t" + str(errorcount)
            #if errorcount/(end - start) >= 0.00001:
                #errorcount50string = errorcount50string + "\n" + chrom + "\t" + str(start) + "\t" + str(end) + "\t" + str(errorcount)
#
        #errorcount30obj = pybedtools.BedTool(errorcount30string, from_string = True)
        #errorcount30obj.saveas(nostrsubq30combreferrorfile)
        #errorcount40obj = pybedtools.BedTool(errorcount40string, from_string = True)
        #errorcount40obj.saveas(nostrsubq40combreferrorfile)
        #errorcount50obj = pybedtools.BedTool(errorcount50string, from_string = True)
        #errorcount50obj.saveas(nostrsubq50combreferrorfile)
#
        #bedtools subtract -a $REFGENOMEBED -b $NLOCBED | bedtools subtract -a - -b $NOSTRSUBQ30BED | bedtools subtract -a - -b $UNCOVBED > $CLEANQ30BED
        #bedtools subtract -a $REFGENOMEBED -b $NLOCBED | bedtools subtract -a - -b $NOSTRSUBQ40BED | bedtools subtract -a - -b $UNCOVBED > $CLEANQ40BED
        #bedtools subtract -a $REFGENOMEBED -b $NLOCBED | bedtools subtract -a - -b $NOSTRSUBQ50BED | bedtools subtract -a - -b $UNCOVBED > $CLEANQ50BED

    return 0
